(*
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2016)
License: LGPL (see file COPYING.LESSER)
*)

(* AL94 (doi:10.1007/3-540-58216-9_40) *)
text \<open>A formalization of the main results from AL94\<close>
theory ALS94
imports
  Conditional_Critical_Pairs
  Conditional_Rewriting
  "../Rewriting/Multihole_Context"
  "../Rewriting/Abstract_Matching"
begin

section \<open>Contextual Rewriting\<close>

text \<open>skol replaces a variable `Var x` by a constant `Fun x []` if x is in set V\<close>
fun skol :: "'v set \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f + 'v, 'v) term"
where
  "skol V (Var x) = (if x \<in> V then Fun (Inr x) [] else Var x)"
| "skol V (Fun f ts) = Fun (Inl f) (map (skol V) ts)"

fun skol_ctxt :: "'v set \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f + 'v, 'v) ctxt"
where
  "skol_ctxt V Hole = Hole"
| "skol_ctxt V (More f us D vs) = More (Inl f) (map (skol V) us) (skol_ctxt V D) (map (skol V) vs)"

lemma skol_ctxt_skol:
  "(skol_ctxt V D)\<langle>skol V t\<rangle> = skol V D\<langle>t\<rangle>"
by (induct D rule: skol_ctxt.induct) auto

fun deskol :: "('f + 'v, 'v) term \<Rightarrow> ('f ,'v) term" where
  "deskol (Var x) = Var x"
| "deskol (Fun (Inr x) ts) = Var x"
| "deskol (Fun (Inl f) ts) = Fun f (map deskol ts)"

lemma deskol_skol [simp]:
  "deskol (skol V t) = t"
by (induct rule: skol.induct) (auto simp: map_idI)

definition skol_trs :: "('f, 'v) trs \<Rightarrow> ('f + 'v, 'v) trs"
where
  "skol_trs C = (\<lambda>c. (skol (vars_trs C) (fst c), skol (vars_trs C) (snd c))) ` C"

text \<open>contextual rewrite step. Two terms `s` and `t` are in `ctxt_step R C` if they are in relation
R or there is an rstep in `skol_trs C` from `skol (vars_trs C) s` to
`skol (vars_trs C) t`.\<close>
inductive_set ctxt_step :: "('f, 'v) term rel \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) term rel" for R C
where
  base: "(s, t) \<in> R \<Longrightarrow> (s, t) \<in> ctxt_step R C"
| skol: "(skol (vars_trs C) s, skol (vars_trs C) t) \<in> rstep (skol_trs C) \<Longrightarrow>
    (s, t) \<in> ctxt_step R C"

lemma conds_sat_iff':
  "(\<exists>n. conds_sat R n C \<sigma>) \<longleftrightarrow> (\<forall>(s, t) \<in> set C. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*)" (is "?L \<longleftrightarrow> ?R")
proof
  assume "?L"
  then show "?R" by (auto simp: conds_sat_iff csteps_n_imp_csteps)
next
  assume "?R"
  { fix x
    assume "x \<in> set C"
    with \<open>?R\<close> have "(fst x \<cdot> \<sigma>, snd x \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" by auto
    from csteps_imp_csteps_n [OF this] have "\<exists>n. (fst x \<cdot> \<sigma>, snd x \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*" .
  }
  then have "\<forall>x \<in> set C. \<exists>n. (fst x \<cdot> \<sigma>, snd x \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*" by auto
  with bchoice [of "set C" "\<lambda>x n. (fst x \<cdot> \<sigma>, snd x \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"]
    obtain f where f: "\<forall>(s, t) \<in> set C. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep_n R (f (s, t)))\<^sup>*" by fastforce   
  moreover obtain n where n: "Max (set (map f C)) = n" (is "Max ?N = _") by blast
  { fix x
    assume a: "x \<in> set C"
    then have x: "f x \<in> ?N" by force
    from f a have 1: "(fst x \<cdot> \<sigma>, snd x \<cdot> \<sigma>) \<in> (cstep_n R (f x))\<^sup>*" by auto
    have "finite ?N" by simp
    from Max_ge [OF this x]  have "f x \<le> n" unfolding n [symmetric] .
    from rtrancl_mono [OF cstep_n_mono [OF this]] 1 have "(fst x \<cdot> \<sigma>, snd x \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"
      by auto
  }
  then show "?L" unfolding conds_sat_iff by fastforce
qed

lemma skol_inj:
  "skol V s = skol V t \<Longrightarrow> s = t"
by (metis deskol_skol)

lemma in_skol_trs_iff:
  "(s, t) \<in> skol_trs C \<longleftrightarrow> (\<exists>(l, r) \<in> C. s = skol (vars_trs C) l \<and> t = skol (vars_trs C) r)"
by (auto simp: skol_trs_def rev_image_eqI)

lemma ground_skol_mono:
  assumes "ground (skol V t)"
    and "V \<subseteq> U"
  shows "ground (skol U t)"
using assms
apply (induct t rule: skol.induct)
apply auto
using set_rev_mp by fastforce

lemma ground_skol_vars_term:
  "ground (skol (vars_term t) t)"
proof (induct t)
  case (Fun f ts)
  let ?t = "Fun f ts"
  { fix i
    assume "i < length ts"
    then have "ts ! i \<in> set ts" by auto
    then have "ground (skol (vars_term (ts ! i)) (ts ! i))" using Fun by auto
  }
  then have "\<forall>i < length ts. ground (skol (vars_term (ts ! i)) (ts ! i))" by auto
  moreover have "\<forall>i < length ts. vars_term (ts ! i) \<subseteq> vars_term ?t" by fastforce
  ultimately have "\<forall>i < length ts. ground (skol (vars_term ?t) (ts ! i))" using ground_skol_mono
    by fastforce
  then have "ground (Fun (Inl f) (map (skol (vars_term ?t)) ts))"
    using in_set_idx by force 
  then show ?case by fastforce
qed (auto)

lemma in_skol_trs_ground:
  assumes "(l, r) \<in> skol_trs C"
  shows "ground l \<and> ground r"
proof -
  from assms obtain s t where s: "skol (vars_trs C) s = l" and t: "skol (vars_trs C) t = r"
    and rule: "(s, t) \<in> C"
    by (auto simp: skol_trs_def)
  have "ground (skol (vars_term s) s)"
    and "ground (skol (vars_term t) t)" using ground_skol_vars_term by auto
  moreover have "vars_term t \<subseteq> vars_trs C"
    and "vars_term s \<subseteq> vars_trs C" using rule
    by (auto simp: vars_trs_def vars_rule_def)
  ultimately show ?thesis using ground_skol_mono
  using s t by blast
qed

lemma skol_skol_ctxt:
  assumes "skol V t = D\<langle>skol V u\<rangle>"
  shows "\<exists>D'. t = D'\<langle>u\<rangle> \<and> D = skol_ctxt V D'"
using assms
proof (induct D arbitrary: V t)
  case (Hole)
  then show ?case by (auto dest!: skol_inj intro: exI [of _ \<box>])
next
  case (More f us F vs)
  then show ?case
  proof (cases t)
    case (Var x)
    then show ?thesis using More by (auto split: if_splits)
  next
    case (Fun g ts)
    from More(2) [unfolded Fun] have ts: "map (skol V) ts = us @ F\<langle>skol V u\<rangle> # vs"
      and f: "f = Inl g" by auto
    from ts obtain us' vs' z
      where us: "us = map (skol V) us'"
      and z: "F\<langle>skol V u\<rangle> = skol V z"
      and vs: "vs = map (skol V) vs'"
      apply (induct us arbitrary: ts)
      by (auto) (metis list.map(2))+
    from More(1) [OF z [symmetric]] obtain F' where "z = F'\<langle>u\<rangle>" and "F = skol_ctxt V F'" by auto
    then have *: "skol V (Fun g ts) =
      (More (Inl g) (map (skol V) us') (skol_ctxt V F') (map (skol V) vs'))\<langle>skol V u\<rangle>"
      (is "_ = ?D\<langle>_\<rangle>") using Fun More.prems f us vs  by blast
    then have "Fun g ts = (More g us' F' vs')\<langle>u\<rangle>" (is "_ = ?D'\<langle>_\<rangle>")
      by (metis skol_ctxt.simps(2) skol_ctxt_skol skol_inj) 
    then have "t = ?D'\<langle>u\<rangle>" by (auto simp: Fun)
    moreover from More(2) [unfolded Fun] * have "More f us F vs =
      More (Inl g) (map (skol V) us') (skol_ctxt V F') (map (skol V) vs')"
      using \<open>F = skol_ctxt V F'\<close> f us vs by blast
    ultimately show ?thesis by fastforce
  qed
qed

lemma ctxt_step_csteps:
  assumes "(s, t) \<in> ctxt_step (cstep R) C"
    and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> C. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
proof -
  { fix s t
    assume 1: "(skol (vars_trs C) s, skol (vars_trs C) t) \<in> rstep (skol_trs C)"
      and 2: "\<forall>x \<in> C. case x of (s\<^sub>i, t\<^sub>i) \<Rightarrow> (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
    from rstepE[OF 1] obtain  D \<tau> l r
      where rule: "(l, r) \<in> skol_trs C"
      and s: "skol (vars_trs C) s = D\<langle>l \<cdot> \<tau>\<rangle>"
      and t: "skol (vars_trs C) t = D\<langle>r \<cdot> \<tau>\<rangle>"
      by metis
    from in_skol_trs_iff [THEN iffD1, OF rule] obtain l' r'
      where l': "l = skol (vars_trs C) l'"
      and r': "r = skol (vars_trs C) r'" by auto
    from in_skol_trs_ground [OF rule] have "ground l" and "ground r" by auto
    then have 3: "skol (vars_trs C) s = D\<langle>l\<rangle>"
      and 4: "skol (vars_trs C) t = D\<langle>r\<rangle>"
      using s t ground_subst_apply by auto
    from 3 [unfolded l'] obtain D' where "skol_ctxt (vars_trs C) D' = D"
      using skol_skol_ctxt by blast
    then have 5: "s = D'\<langle>l'\<rangle>" and 6: "t = D'\<langle>r'\<rangle>" by (metis 3 4 l' r' skol_ctxt_skol skol_inj)+
    have "(l', r') \<in> C" using rule [unfolded l' r']
      by (auto simp: skol_trs_def) (metis deskol_skol)
    with 2 have "(l' \<cdot> \<sigma>, r' \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" by auto
    from csteps_ctxt [OF this, of "D' \<cdot>\<^sub>c \<sigma>"]
      have "((D' \<cdot>\<^sub>c \<sigma>)\<langle>l' \<cdot> \<sigma>\<rangle>, (D' \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<sigma>\<rangle>) \<in> (cstep R)\<^sup>*" .
    then have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" unfolding 5 6 by fastforce
  }
  with assms show ?thesis
    by (induct rule: ctxt_step.induct) (auto dest: cstep_subst)
qed

text \<open>Lemma 4.2 from AL94\<close>
lemma ctxt_steps_csteps:
  assumes "(s, t) \<in> (ctxt_step (cstep R) C)\<^sup>*"
    and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> C. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
using assms(1)
proof (induct rule: rtrancl_induct)
  case (step u v)
  from ctxt_step_csteps [OF step(2) assms(2)] have "(u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" .
  then show ?case using step(1,3) by auto
qed (auto)

text \<open>Definition 4.1 from AL94\<close>
definition strongly_deterministic :: "('f, 'v) ctrs \<Rightarrow> bool"
where
  "strongly_deterministic R \<longleftrightarrow> (\<forall> ((l, r), cs) \<in> R. \<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs. strongly_irreducible R t\<^sub>i)"

text \<open>Definition 4.1 from AL94\<close>
definition sdtrs :: "('f, 'v) ctrs \<Rightarrow> bool"
where
  "sdtrs R \<longleftrightarrow> strongly_deterministic R \<and> wf_ctrs R"

definition all_CPs_joinable
where
  "all_CPs_joinable R = (\<forall>s t cs. ((s, t), cs) \<in> CCP R \<longrightarrow>
    (\<forall>\<sigma>. (\<exists>n. conds_sat R n cs \<sigma>) \<longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>\<down>))"

(* TODO move to Conditional_Rewriting.thy (?) *)
lemma variant_conds_sat_cstep_n:
  assumes "\<pi> \<bullet> \<rho> \<in> R"
    and "conds_sat R n (conds \<rho>) \<sigma>"
  shows "(clhs \<rho> \<cdot> \<sigma>, crhs \<rho> \<cdot> \<sigma>) \<in> cstep_n R (Suc n)"
proof -
  from assms(1) have r: "((\<pi> \<bullet> clhs \<rho>, \<pi> \<bullet> crhs \<rho>), \<pi> \<bullet> conds \<rho>) \<in> R"
    by (simp add: eqvt split: prod.split)
       (metis crule_pt.snd_eqvt prod.collapse rules_pt.permute_list_def)
  from assms(2) have "conds_sat R n (- \<pi> \<bullet> \<pi> \<bullet> conds \<rho>) \<sigma>"
    by (metis rules_pt.permute_minus_cancel(2))
  from conds_sat_perm_shift [THEN iffD1, OF this]
    have "conds_sat R n (\<pi> \<bullet> conds \<rho>) (\<sigma> \<circ> Rep_perm (- \<pi>))" (is "conds_sat _ _ _ ?\<tau>") by auto
  from conds_sat_iff [THEN iffD1, OF this]
    have "\<forall>(s\<^sub>i, t\<^sub>i)\<in>set (\<pi> \<bullet> conds \<rho>). (s\<^sub>i \<cdot> ?\<tau>, t\<^sub>i \<cdot> ?\<tau>) \<in> (cstep_n R n)\<^sup>*" .
  from cstep_n_SucI [OF r this, of _ \<box>]
    have "(\<pi> \<bullet> clhs \<rho> \<cdot> ?\<tau>, \<pi> \<bullet> crhs \<rho> \<cdot> ?\<tau>) \<in> cstep_n R (Suc n)" by force
  then have "(clhs \<rho> \<cdot> (sop \<pi> \<circ>\<^sub>s ?\<tau>), crhs \<rho> \<cdot> (sop \<pi> \<circ>\<^sub>s ?\<tau>)) \<in> cstep_n R (Suc n)" by simp
  then show ?thesis
    by (metis (no_types, lifting) permute_term_subst_apply_term subst_subst
        term_apply_subst_Var_Rep_perm term_pt.permute_minus_cancel(2))
qed

(* TODO: move to Conditional_Critical_Pairs.thy (?) *)
lemma CCP_E:
  fixes s :: "('f, 'v :: infinite) term"
  assumes "((s, t), cs) \<in> CCP R"
  shows "\<exists>\<rho>\<^sub>1 \<rho>\<^sub>2 p \<sigma>. s = crhs \<rho>\<^sub>1 \<cdot> \<sigma> \<and> t = replace_at (clhs \<rho>\<^sub>1) p (crhs \<rho>\<^sub>2) \<cdot> \<sigma> \<and>
    cs = subst_list \<sigma> (snd \<rho>\<^sub>1 @ snd \<rho>\<^sub>2) \<and>
    overlap R \<rho>\<^sub>1 \<rho>\<^sub>2 p \<and> mgu (clhs \<rho>\<^sub>1 |_ p) (clhs \<rho>\<^sub>2) = Some \<sigma> \<and> (p = \<epsilon> \<longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2))"
using assms unfolding CCP_def by blast

text \<open>easy direction of Theorem 4.1 from AL94 (no need for strong determinism nor quasi-reductivity)\<close>
lemma CR_imp_all_CPs_joinable:
  assumes "CR (cstep R)"
  shows "all_CPs_joinable R"
unfolding all_CPs_joinable_def
proof (intro allI impI)
  fix s t cs \<sigma>
  assume ccp: "((s, t), cs) \<in> CCP R"
    and csat: "\<exists>n. conds_sat R n cs \<sigma>"
  from CCP_E [OF ccp] obtain \<rho>\<^sub>1 \<rho>\<^sub>2 p \<mu> where s: "s = crhs \<rho>\<^sub>1 \<cdot> \<mu>"
    and t: "t = (ctxt_of_pos_term p (clhs \<rho>\<^sub>1))\<langle>crhs \<rho>\<^sub>2\<rangle> \<cdot> \<mu>"
    and cs: "cs = subst_list \<mu> (conds \<rho>\<^sub>1 @ conds \<rho>\<^sub>2)"
    and o: "overlap R \<rho>\<^sub>1 \<rho>\<^sub>2 p"
    and \<mu>: "mgu (clhs \<rho>\<^sub>1 |_ p) (clhs \<rho>\<^sub>2) = Some \<mu>"
    and "p = \<epsilon> \<longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2)" by auto
  from o have p: "p \<in> poss (clhs \<rho>\<^sub>1)" by (auto simp: overlap_def intro: funposs_imp_poss)
  from csat obtain n where conds: "conds_sat R n (conds \<rho>\<^sub>1 @ conds \<rho>\<^sub>2) (\<mu> \<circ>\<^sub>s \<sigma>)"
    (is "conds_sat _ _ (?C\<^sub>1 @ ?C\<^sub>2) ?\<tau>") unfolding cs using conds_sat_subst_list by blast
  from conds_sat_mono' [OF conds, of ?C\<^sub>1] have c1: "conds_sat R n (conds \<rho>\<^sub>1) ?\<tau>" by force
  from conds_sat_mono' [OF conds, of ?C\<^sub>2] have c2: "conds_sat R n (conds \<rho>\<^sub>2) ?\<tau>" by force
  obtain q q' where r1: "q \<bullet> \<rho>\<^sub>1 \<in> R" and r2: "q' \<bullet> \<rho>\<^sub>2 \<in> R" using o by (auto simp: overlap_def)
  from variant_conds_sat_cstep_n [OF r1 c1]
    have "(clhs \<rho>\<^sub>1 \<cdot> ?\<tau>, crhs \<rho>\<^sub>1 \<cdot> ?\<tau>) \<in> cstep_n R (Suc n)" .
  then have "(clhs \<rho>\<^sub>1 \<cdot> ?\<tau>, crhs \<rho>\<^sub>1 \<cdot> ?\<tau>) \<in> cstep R" using cstep_iff by blast
  then have 2: "(clhs \<rho>\<^sub>1 \<cdot> ?\<tau>, crhs \<rho>\<^sub>1 \<cdot> ?\<tau>) \<in> (cstep R)\<^sup>*" by simp
  from variant_conds_sat_cstep_n [OF r2 c2]
    have "(clhs \<rho>\<^sub>2 \<cdot> ?\<tau>, crhs \<rho>\<^sub>2 \<cdot> ?\<tau>) \<in> cstep_n R (Suc n)" .
  then have "(clhs \<rho>\<^sub>2 \<cdot> ?\<tau>, crhs \<rho>\<^sub>2 \<cdot> ?\<tau>) \<in> cstep R" using cstep_iff by blast
  then have "(replace_at (clhs \<rho>\<^sub>1) p (clhs \<rho>\<^sub>2) \<cdot> ?\<tau>, replace_at (clhs \<rho>\<^sub>1) p (crhs \<rho>\<^sub>2) \<cdot> ?\<tau>) \<in> cstep R"
    by (simp add: cstep_ctxt)
  then have 0: "(replace_at (clhs \<rho>\<^sub>1) p (clhs \<rho>\<^sub>2) \<cdot> ?\<tau>, replace_at (clhs \<rho>\<^sub>1) p (crhs \<rho>\<^sub>2) \<cdot> ?\<tau>) \<in> (cstep R)\<^sup>*"
    by simp
  have "clhs \<rho>\<^sub>1 |_ p \<cdot> \<mu> = clhs \<rho>\<^sub>2 \<cdot> \<mu>"
    using mgu_sound [OF \<mu>] by (auto simp: is_imgu_def unifiers_def)
  then have e: "clhs \<rho>\<^sub>1 \<cdot> \<mu> |_ p = clhs \<rho>\<^sub>2 \<cdot> \<mu>" by (simp add: p)
  have "p \<in> poss (clhs \<rho>\<^sub>1 \<cdot> \<mu>)" using p by fastforce
  from replace_at_ident [OF this e]
    have "replace_at (clhs \<rho>\<^sub>1) p (clhs \<rho>\<^sub>2) \<cdot> \<mu> = clhs \<rho>\<^sub>1 \<cdot> \<mu>"
    by (simp add: ctxt_of_pos_term_subst p) 
  then have "replace_at (clhs \<rho>\<^sub>1) p (clhs \<rho>\<^sub>2) \<cdot> ?\<tau> = clhs \<rho>\<^sub>1 \<cdot> ?\<tau>" by simp
  with 0 have 1: "(clhs \<rho>\<^sub>1 \<cdot> ?\<tau>, replace_at (clhs \<rho>\<^sub>1) p (crhs \<rho>\<^sub>2) \<cdot> ?\<tau>) \<in> (cstep R)\<^sup>*" by auto
  from 1 2 have "(crhs \<rho>\<^sub>1 \<cdot> ?\<tau>, replace_at (clhs \<rho>\<^sub>1) p (crhs \<rho>\<^sub>2) \<cdot> ?\<tau>) \<in> (cstep R)\<^sup>\<down>"
    using assms [unfolded CR_on_def] by blast
  then show "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>\<down>" unfolding s t by auto
qed

(* TODO: move to Conditional_Rewriting.thy (?) *)
lemma quasi_reductive_order_csteps:
  assumes "quasi_reductive_order R S"
  shows "(cstep R)\<^sup>* \<subseteq> S\<^sup>="
proof -
  from assms have trans: "trans S" and RsubS: "cstep R \<subseteq> S"
    using quasi_reductive_order_cstep [OF assms] by (auto simp: quasi_reductive_order_def)
  have "(cstep R)\<^sup>* \<subseteq> S^*" by (rule rtrancl_mono[OF RsubS])
  also have "S^* = (S^+)^=" by regexp
  also have "S^+ = S" using trans by auto
  finally show ?thesis by blast
qed

text \<open>Corresponds to one half of Corollary 3.1 (or 3.3 as it is referred to later) from AL94\<close>
lemma quasi_reductive_order_conds_sat:
  assumes qr: "quasi_reductive_order R S"
    and rule: "((l, r), cs) \<in> R"
    and cs: "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
  shows "\<forall>i < length cs. (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> (S \<union> {\<rhd>})\<^sup>+"
proof -
  from quasi_reductive_order_csteps [OF qr] have "(cstep R)\<^sup>* \<subseteq> S\<^sup>=" .
  moreover from qr [unfolded quasi_reductive_order_def]
    have *: "\<forall>l r cs \<sigma>. ((l, r), cs) \<in> R \<longrightarrow>
    (\<forall>i < length cs. (\<forall>j <i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S\<^sup>=) \<longrightarrow>
    (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> (S \<union> {\<rhd>})\<^sup>+)" by blast
  ultimately have "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> S\<^sup>=" using cs by blast
  then have "\<forall>i < length cs. \<forall>j<i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S\<^sup>=" by force
  with * rule show ?thesis by auto
qed

(* TODO: move to Conditional_Rewriting.thy *)
lemma quasi_reductive_obtains_normalized_subst:
  assumes "quasi_reductive_order R S"
  obtains \<tau> where "normalized R \<tau>" and "\<forall>x. (\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*"
    and "\<tau> = (\<lambda>x. some_NF (cstep R) (\<sigma> x))"
proof -
  have *: "SN (cstep R)" using assms using quasi_reductive_SN quasi_reductive_def by auto
  def \<tau> \<equiv> "\<lambda>x. some_NF (cstep R) (\<sigma> x)"
  { fix x
    from * have "(\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*" unfolding \<tau>_def by (simp add: SN_def some_NF)
    then have "(\<sigma>  x, \<tau> x) \<in> (cstep R)\<^sup>*" by auto
  }
  then have "\<forall>x. (\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*" by auto
  moreover have "normalized R \<tau>" by (auto simp: normalized_def) (metis SN_def * \<tau>_def some_NF)
  ultimately obtain \<tau> where "normalized R \<tau>" and  "\<forall>x. (\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*"
    and "\<tau> = (\<lambda>x. some_NF (cstep R) (\<sigma> x))" using \<tau>_def by blast
  then show ?thesis ..
qed

(* TODO: move to Conditional_Rewriting.thy *)
lemma quasi_decreasing_order_csteps:
  assumes "quasi_decreasing_order R S"
  shows "(cstep R)\<^sup>* \<subseteq> S\<^sup>="
proof -
  from assms have trans: "trans S" and RsubS: "cstep R \<subseteq> S"
    by (auto simp: quasi_decreasing_order_def)
  have "(cstep R)\<^sup>* \<subseteq> S^*" by (rule rtrancl_mono[OF RsubS])
  also have "S^* = (S^+)^=" by regexp
  also have "S^+ = S" using trans by auto
  finally show ?thesis by blast
qed

(* TODO: move to Conditional_Rewriting.thy (?) *)
lemma quasi_decreasing_order_conds_sat:
  assumes qd: "quasi_decreasing_order R S"
    and rule: "((l, r), cs) \<in> R"
    and cs: "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
  shows "\<forall>i < length cs. (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> (S \<union> {\<rhd>})\<^sup>+"
proof -
  from quasi_decreasing_order_csteps [OF qd] have "(cstep R)\<^sup>* \<subseteq> S\<^sup>=" .
  moreover from qd [unfolded quasi_decreasing_order_def]
    have *: "\<forall>l r cs \<sigma>. ((l, r), cs) \<in> R \<longrightarrow>
    (\<forall>i < length cs. (\<forall>j < i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*) \<longrightarrow>
    (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> (S \<union> {\<rhd>})\<^sup>+)" by blast
  ultimately have "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" using cs by blast
  then have "\<forall>i < length cs. \<forall>j<i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" by force
  with * rule show ?thesis by auto
qed

(* TODO: move to Conditional_Rewriting.thy (?) *)
lemma quasi_decreasing_obtains_normalized_subst:
  assumes "quasi_decreasing_order R S"
  obtains \<tau> where "normalized R \<tau>" and "\<forall>x. (\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*"
    and "\<tau> = (\<lambda>x. some_NF (cstep R) (\<sigma> x))"
proof -
  have *: "SN (cstep R)" using assms using quasi_decreasing_SN quasi_decreasing_def by auto
  def \<tau> \<equiv> "\<lambda>x. some_NF (cstep R) (\<sigma> x)"
  { fix x
    from * have "(\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*" unfolding \<tau>_def by (simp add: SN_def some_NF)
    then have "(\<sigma>  x, \<tau> x) \<in> (cstep R)\<^sup>*" by auto
  }
  then have "\<forall>x. (\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*" by auto
  moreover have "normalized R \<tau>" by (auto simp: normalized_def) (metis SN_def * \<tau>_def some_NF)
  ultimately obtain \<tau> where "normalized R \<tau>" and  "\<forall>x. (\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*"
    and "\<tau> = (\<lambda>x. some_NF (cstep R) (\<sigma> x))" using \<tau>_def by blast
  then show ?thesis ..
qed

lemma aux_lemma:
  fixes l\<^sub>1 :: "('f, 'v :: infinite) term"
  assumes qro: "quasi_decreasing_order R S"
    and allCPs: "all_CPs_joinable R"
    and SDTRS: "sdtrs R"
    and r1: "((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R" (is "?\<rho>\<^sub>1 \<in> _")
    and cs1: "\<forall>(s, t) \<in> set cs\<^sub>1. (s \<cdot> \<sigma>\<^sub>1, t \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
    and C: "t = C\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>"
    and C': "t\<^sub>1 = C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>"
    and r2: "((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R" (is "?\<rho>\<^sub>2 \<in> _")
    and cs2: "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
    and D: "t = D\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
    and D': "t\<^sub>2 = D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
    and q: "hole_pos C = q"
    and p: "hole_pos D = p"
    and GE: "q \<le> p"
    and IH: "\<And>y t' t''. (y, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse> \<Longrightarrow> (y, t') \<in> (cstep R)\<^sup>*
      \<Longrightarrow> (y, t'') \<in> (cstep R)\<^sup>* \<Longrightarrow> (t', t'') \<in> (cstep R)\<^sup>\<down>"
  shows "(t\<^sub>1, t\<^sub>2) \<in> (cstep R)\<^sup>\<down>"
proof -
  have tq': "t |_q = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using C q by fastforce
  from cstepI [OF r1 cs1, of "l\<^sub>1 \<cdot> \<sigma>\<^sub>1" \<box> "r\<^sub>1 \<cdot> \<sigma>\<^sub>1"] have s1: "(l\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> cstep R" by simp
  from cstepI [OF r2 cs2, of "l\<^sub>2 \<cdot> \<sigma>\<^sub>2" \<box> "r\<^sub>2 \<cdot> \<sigma>\<^sub>2"] have s2: "(l\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> cstep R" by simp
  from GE have r: "p = q <#> pos_diff p q" (is "_ = _ <#> ?r") by simp
  consider (ST) "t \<rhd> t |_q" | (EQ) "t = t |_q"
    using C hole_pos_poss q subt_at_imp_supteq by blast
  then show ?thesis
  proof (cases)
    case (ST)
    from GE ST have st: "t |_q \<unrhd> t |_p" using D hole_pos_poss less_eq_pos_imp_supt_eq p by blast 
    have *: "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 = replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) ?r (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)"
    proof -
      have "hole_pos (D |_c q) = pos_diff p q" using r hole_pos_subt_at_ctxt p by blast
      then show ?thesis by (metis D r tq' ctxt_of_pos_term_hole_pos p subt_at_subt_at_ctxt)
    qed
    have tq: "(t |_q, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" using ST by blast
    have 1: "(t |_q, r\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*" unfolding tq' using s1 by blast
    have 2: "(t |_q, replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) ?r (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) \<in> (cstep R)\<^sup>*" (is "(_, ?t2) \<in> _")
      unfolding tq' by (metis * csteps_ctxt r_into_rtrancl s2)
    from IH [OF tq 1 2] have j: "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, ?t2) \<in> (cstep R)\<^sup>\<down>" .
    then obtain s where "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, s) \<in> (cstep R)\<^sup>*" and "(?t2, s) \<in> (cstep R)\<^sup>*" by blast
    then have 5: "(replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1), replace_at t q s) \<in> (cstep R)\<^sup>*"
      and 6: "(replace_at t q ?t2, replace_at t q s) \<in> (cstep R)\<^sup>*"
      by (simp add: csteps_ctxt)+
    have 3: "t\<^sub>1 = replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using C C' q by auto
    have 4: "t\<^sub>2 = replace_at t q ?t2" using r
      by (metis C D D' ctxt_ctxt ctxt_of_pos_term_append
          ctxt_of_pos_term_hole_pos hole_pos_poss p q tq')
    show ?thesis using 5 6 unfolding 3 4 by blast
  next
    case (EQ)
    have t: "t = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using EQ tq' by simp
    have tp': "t |_ p = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using D p by fastforce
    consider (\<alpha>) "p \<in> funposs l\<^sub>1" | (\<beta>) "p \<notin> funposs l\<^sub>1" by auto
    then show ?thesis
    proof (cases)
      case (\<alpha>)
      from vars_crule_disjoint obtain \<pi>
        where \<pi>: "vars_crule (\<pi> \<bullet> ?\<rho>\<^sub>1) \<inter> vars_crule ?\<rho>\<^sub>2 = {}" ..
      def l \<equiv> "\<pi> \<bullet> l\<^sub>1" and r \<equiv> "\<pi> \<bullet> r\<^sub>1" and cs \<equiv> "\<pi> \<bullet> cs\<^sub>1" and \<sigma> \<equiv> "sop (-\<pi>) \<circ>\<^sub>s \<sigma>\<^sub>1"
      note rename = l_def r_def cs_def \<sigma>_def
      
      have *: "-\<pi> \<bullet> ((l, r), cs) \<in> R" (is "-\<pi> \<bullet> ?\<rho> \<in> R") and "0 \<bullet> ?\<rho>\<^sub>2 \<in> R"
        using r1 r2 by (simp_all add: eqvt rename o_def)
      then have rule_variants: "\<exists>\<pi>. \<pi> \<bullet> ?\<rho> \<in> R" "\<exists>\<pi>. \<pi> \<bullet> ?\<rho>\<^sub>2 \<in> R"
        by blast+
  
      have disj: "vars_crule ?\<rho> \<inter> vars_crule ?\<rho>\<^sub>2 = {}"
        using \<pi> by (auto simp: eqvt rename)
      
      from cs1 have cs: "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
        by (auto simp: rename rule_pt.permute_prod.simps)
        
      from \<alpha> have funposs: "p \<in> funposs (clhs ?\<rho>)" by (simp add: rename)
      then have poss: "p \<in> poss l" by (auto dest: funposs_imp_poss)
      with vars_term_subt_at [OF this] and disj
        have "vars_term (l |_ p) \<inter> vars_term l\<^sub>2 = {}" and "(l |_ p) \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
        apply (auto simp: eqvt vars_crule_def vars_rule_def rename)
        using EQ tp' tq' by auto
      from vars_term_disjoint_imp_unifier [OF this]
        obtain \<tau> where mgu: "mgu (l |_ p) l\<^sub>2 = Some \<tau>"
        using mgu_complete by (auto simp: unifiers_def)
      consider (improp) "p = \<epsilon> \<and> (\<exists>\<pi>. \<pi> \<bullet> ?\<rho> = ?\<rho>\<^sub>2)"
        | (proper) "\<not>p = \<epsilon> \<or> \<not>(\<exists>\<pi>. \<pi> \<bullet> ?\<rho> = ?\<rho>\<^sub>2)" by blast
      then show ?thesis
      proof (cases)
        case (improp)
        from improp have t': "t = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using t by blast
        from improp have t: "t = l \<cdot> \<sigma>" "t = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
          using \<open>l |_ p \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> subt_at.simps(1) tp' by auto
        from improp obtain \<pi>' where "\<pi>' \<bullet> ?\<rho> = ?\<rho>\<^sub>2" by auto
        then have ren: "(\<pi>' + \<pi>) \<bullet> ?\<rho>\<^sub>1 = ?\<rho>\<^sub>2"
          (is "?\<pi> \<bullet> ((_, _), _) = _") using rename by (auto simp: eqvt)
        then have ren': "l\<^sub>2 = ?\<pi> \<bullet> l\<^sub>1" "r\<^sub>2 = ?\<pi> \<bullet> r\<^sub>1" "cs\<^sub>2 = ?\<pi> \<bullet> cs\<^sub>1" by (auto simp: eqvt)
        have ren'': "l\<^sub>1 = -?\<pi> \<bullet> l\<^sub>2" "r\<^sub>1 = -?\<pi> \<bullet> r\<^sub>2" "r\<^sub>1 = -?\<pi> \<bullet> r\<^sub>2"
          by (auto simp: ren') (metis term_pt.permute_minus_cancel(2) term_pt.permute_plus)+
        have ren_cs: "cs\<^sub>1 = -?\<pi> \<bullet> cs\<^sub>2" using ren' by (metis rules_pt.permute_minus_cancel(2))
        have l2subst: "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = l\<^sub>2 \<cdot> (sop (-?\<pi>) \<circ>\<^sub>s \<sigma>\<^sub>1)" (is "_ = _ \<cdot> ?\<sigma>\<^sub>1")
          using t t' ren'' by (auto simp: eqvt)
        from this [symmetric] have substl: "\<forall>x \<in> vars_term l\<^sub>2. ?\<sigma>\<^sub>1 x = \<sigma>\<^sub>2 x"
          using term_subst_eq_rev by blast
  
        have "t = l\<^sub>2 \<cdot> ?\<sigma>\<^sub>1" using \<open>l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = l\<^sub>2 \<cdot> sop (- (\<pi>' + \<pi>)) \<circ>\<^sub>s \<sigma>\<^sub>1\<close> t(2) by blast
        have t1': "t\<^sub>1 = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" by (metis C C' EQ ctxt_apply_term.simps(1) ctxt_supt supt_neqD tq')
        from t1' have t1: "t\<^sub>1 = r\<^sub>2 \<cdot> ?\<sigma>\<^sub>1" using ren'' by simp
        from improp have t2: "t\<^sub>2 = r\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (metis D' p subt_at.simps(1) subt_at_hole_pos)

        obtain \<sigma>\<^sub>1' where norm_\<sigma>\<^sub>1': "normalized R \<sigma>\<^sub>1'" and \<sigma>\<^sub>1': "\<forall>x. (?\<sigma>\<^sub>1 x, \<sigma>\<^sub>1' x) \<in> (cstep R)\<^sup>*"
          and \<sigma>\<^sub>1'_def: "\<sigma>\<^sub>1' = (\<lambda>x. some_NF (cstep R) (?\<sigma>\<^sub>1 x))"
          using quasi_decreasing_obtains_normalized_subst [OF qro] by auto
        obtain \<sigma>\<^sub>2' where norm_\<sigma>\<^sub>2': "normalized R \<sigma>\<^sub>2'" and \<sigma>\<^sub>2': "\<forall>x. (\<sigma>\<^sub>2 x, \<sigma>\<^sub>2' x) \<in> (cstep R)\<^sup>*"
          and \<sigma>\<^sub>2'_def: "\<sigma>\<^sub>2' = (\<lambda>x. some_NF (cstep R) (\<sigma>\<^sub>2 x))"
          using quasi_decreasing_obtains_normalized_subst [OF qro] by auto
        
        have "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
          using cs2 by (auto simp: eqvt)
        then have cs2': "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
          using csteps_n_subset_csteps by force
        have "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> ?\<sigma>\<^sub>1, t \<cdot> ?\<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
          using cs1 [unfolded ren_cs] by (auto simp: eqvt)
        then have cs2'': "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> ?\<sigma>\<^sub>1, t \<cdot> ?\<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
          using csteps_n_subset_csteps by blast

        { fix x
          from quasi_decreasing_order_conds_sat [OF qro r2 cs2'']
            have 1: "\<forall>i < length cs\<^sub>2. (fst (cs\<^sub>2 ! i) \<cdot> ?\<sigma>\<^sub>1, l\<^sub>2 \<cdot> ?\<sigma>\<^sub>1) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" by auto
          from quasi_decreasing_order_conds_sat [OF qro r2 cs2']
            have 2: "\<forall>i < length cs\<^sub>2. (fst (cs\<^sub>2 ! i) \<cdot> \<sigma>\<^sub>2, l\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" by auto
          from SDTRS [unfolded sdtrs_def strongly_deterministic_def] r2
            have 3: "\<forall>(u, v) \<in> set cs\<^sub>2. strongly_irreducible R v" by fast
  
          let ?det = "\<lambda>x. \<forall>i < length x. vars_term (fst (x ! i)) \<subseteq> vars_term l\<^sub>2 \<union> \<Union>(vars_term ` rhss (set (take i x)))"
          assume "x \<in> evars_crule ?\<rho>\<^sub>2"
          moreover from SDTRS [unfolded sdtrs_def wf_ctrs_def dctrs_def X_vars_def] r2
            have "?det cs\<^sub>2" by fastforce
          ultimately have "\<sigma>\<^sub>1' x = \<sigma>\<^sub>2' x" using cs2' cs2'' 1 2 3
          proof (induct cs\<^sub>2 arbitrary: x rule: List.rev_induct)
            case Nil
            then show ?case by (auto simp: evars_crule_def vars_trs_def)
          next
            case (snoc d ds)
            from snoc(3) have det: "?det ds"
              unfolding take_append nth_append by (auto split: if_splits)
            from snoc(4) have sat1: "\<forall>(s,t) \<in> set ds. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
              by fastforce
            from snoc(5) have sat2: "\<forall>(s,t) \<in> set ds. (s \<cdot> ?\<sigma>\<^sub>1, t \<cdot> ?\<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
              by fastforce
            from snoc(6) have qr1: "\<forall>i<length ds. (fst (ds ! i) \<cdot> ?\<sigma>\<^sub>1, l\<^sub>2 \<cdot> ?\<sigma>\<^sub>1) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>"
              unfolding nth_append by (simp split: if_splits add: o_def)
            from snoc(7) have qr2: "\<forall>i<length ds. (fst (ds ! i) \<cdot> \<sigma>\<^sub>2, l\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>"
              unfolding nth_append by (simp split: if_splits add: o_def)
            from snoc(8) have si: "\<forall>(u, v) \<in> set ds. strongly_irreducible R v" by auto
              
            from snoc(1) [OF _ det sat1 sat2 qr1 qr2 si]
              have IH': "\<And>y. y\<in> evars_crule ((l\<^sub>2, r\<^sub>2), ds) \<Longrightarrow> \<sigma>\<^sub>1' y = \<sigma>\<^sub>2' y" by blast
            consider (DS) "x \<in> evars_crule ((l\<^sub>2, r\<^sub>2), ds)"
              | (U) u where "x \<in> vars_term u" "u = fst d"
              | (V) v where "x \<in> vars_term v" "v = snd d"
              using snoc(2) by (auto simp: evars_crule_def vars_trs_def vars_rule_def)
            then show ?case
            proof (cases)
              case (DS)
              from IH' [OF this] show ?thesis .
            next
              case (U)
              then have "x \<in> vars_term (fst ((ds @ [d]) ! (length ds)))" by simp
              with snoc(3) have "x \<in> vars_term l\<^sub>2 \<union> \<Union>(vars_term ` rhss (set (take (length ds) (ds @ [d]))))"
                by fastforce
              with snoc(2) have "x \<in> evars_crule ((l\<^sub>2, r\<^sub>2), ds)"
                by (auto simp: evars_crule_def vars_defs)
              from IH' [OF this] show ?thesis .
            next
              case (V)
              let ?u = "fst d"
              { fix y
                assume "y \<in> vars_term ?u"
                then have "y \<in> vars_term (fst ((ds @ [d]) ! (length ds)))" by simp
                with snoc(3) have "y \<in> vars_term l\<^sub>2 \<union>
                  \<Union>(vars_term ` rhss (set (take (length ds) (ds @ [d]))))"
                  by fastforce
                then consider (L) "y \<in> vars_term l\<^sub>2" | (E) "y \<in> evars_crule ((l\<^sub>2, r\<^sub>2), ds)"
                  by (auto simp: evars_crule_def vars_defs)
                then have "(?\<sigma>\<^sub>1 y, \<sigma>\<^sub>2 y) \<in> (cstep R)\<^sup>\<down>"
                proof (cases)
                  case (L)
                  then have "?\<sigma>\<^sub>1 y = \<sigma>\<^sub>2 y" by (simp add: substl)
                  then show ?thesis by auto
                next
                  case (E)
                  from IH' [OF E] have "\<sigma>\<^sub>1' y = \<sigma>\<^sub>2' y" .
                  then show ?thesis using \<sigma>\<^sub>1' \<sigma>\<^sub>2' joinI by auto
                qed
              }
              then have "\<And>y. y \<in> vars_term ?u \<Longrightarrow> (?\<sigma>\<^sub>1 y, \<sigma>\<^sub>2 y) \<in> (cstep R)\<^sup>\<down>" by auto
              with term_subst_csteps_join have "(?u \<cdot> ?\<sigma>\<^sub>1, ?u \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>\<down>" by blast
              then obtain s\<^sub>0 where u1: "(?u \<cdot> ?\<sigma>\<^sub>1, s\<^sub>0) \<in> (cstep R)\<^sup>*"
                and u2: "(?u \<cdot> \<sigma>\<^sub>2, s\<^sub>0) \<in> (cstep R)\<^sup>*" by blast
              have d': "(fst d, snd d) \<in> set (ds @ [d])" by simp
              then have "(?u \<cdot> ?\<sigma>\<^sub>1, v \<cdot> ?\<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*" unfolding V(2) using snoc(5) by blast
              moreover from substs_csteps [OF \<sigma>\<^sub>1' [rule_format]] have "(v \<cdot> ?\<sigma>\<^sub>1, v \<cdot> \<sigma>\<^sub>1') \<in> (cstep R)\<^sup>*" by fast
              finally have u1': "(?u \<cdot> ?\<sigma>\<^sub>1, v \<cdot> \<sigma>\<^sub>1') \<in> (cstep R)\<^sup>*" by blast
              have "(?u \<cdot> ?\<sigma>\<^sub>1, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" using snoc(6) unfolding \<open>t = l\<^sub>2 \<cdot> ?\<sigma>\<^sub>1\<close>
                by (metis in_set_conv_nth prod.collapse d')
              from IH [OF this u1' u1] have "(v \<cdot> \<sigma>\<^sub>1', s\<^sub>0) \<in> (cstep R)\<^sup>\<down>" .
              then obtain s\<^sub>1 where s1': "(v \<cdot> \<sigma>\<^sub>1', s\<^sub>1) \<in> (cstep R)\<^sup>*" and s1: "(s\<^sub>0, s\<^sub>1) \<in> (cstep R)\<^sup>*" by auto
              have *: "(?u \<cdot> \<sigma>\<^sub>2, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" using snoc(7) unfolding \<open>t = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close>
                by (metis in_set_conv_nth prod.collapse d')
              from u2 s1 have u2': "(fst d \<cdot> \<sigma>\<^sub>2, s\<^sub>1) \<in> (cstep R)\<^sup>*" by auto
              have "(?u \<cdot> \<sigma>\<^sub>2, v \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*" unfolding V(2) using snoc(4) d' by blast
              moreover from substs_csteps [OF \<sigma>\<^sub>2' [rule_format]] have "(v \<cdot> \<sigma>\<^sub>2, v \<cdot> \<sigma>\<^sub>2') \<in> (cstep R)\<^sup>*" by fast
              finally have u2'': "(?u \<cdot> \<sigma>\<^sub>2, v \<cdot> \<sigma>\<^sub>2') \<in> (cstep R)\<^sup>*" by blast
              from IH [OF * u2' u2''] have 2: "(s\<^sub>1, v \<cdot> \<sigma>\<^sub>2') \<in> (cstep R)\<^sup>\<down>" .
              have 5: "(v \<cdot> \<sigma>\<^sub>1', v \<cdot> \<sigma>\<^sub>2') \<in> (cstep R)\<^sup>\<down>" using 2 s1' rtrancl_join_join by metis
              from snoc(8) d' V(2) have siv: "strongly_irreducible R v" by auto
              from norm_\<sigma>\<^sub>1' siv have 3: "v \<cdot> \<sigma>\<^sub>1' \<in> NF (cstep R)" by (auto simp: strongly_irreducible_def)
              from norm_\<sigma>\<^sub>2' siv have 4: "v \<cdot> \<sigma>\<^sub>2' \<in> NF (cstep R)" by (auto simp: strongly_irreducible_def)
              from 3 4 5 show ?thesis by (meson V(1) join_NF_imp_eq term_subst_eq_rev)
            qed
          qed
        }
        then have evars_join: "\<forall>x \<in> evars_crule ?\<rho>\<^sub>2. \<sigma>\<^sub>1' x = \<sigma>\<^sub>2' x" by auto
        from SDTRS [unfolded sdtrs_def wf_ctrs_def type3_def] r2
          have "vars_term r\<^sub>2 \<subseteq> vars_term l\<^sub>2 \<union> vars_trs (set cs\<^sub>2)"
          using fst_conv snd_conv by fastforce
        then have "vars_crule ?\<rho>\<^sub>2 = evars_crule ?\<rho>\<^sub>2 \<union> vars_term l\<^sub>2"
          unfolding evars_crule_def
          proof -
            have "vars_crule ?\<rho>\<^sub>2 = vars_trs (set cs\<^sub>2) \<union> (vars_term l\<^sub>2 \<union> vars_term r\<^sub>2)"
              by (simp add: inf_sup_aci(5) vars_crule_def vars_rule_def)
            then show "vars_crule ?\<rho>\<^sub>2 = vars_trs (set (conds ?\<rho>\<^sub>2)) - vars_term (clhs ?\<rho>\<^sub>2) \<union> vars_term l\<^sub>2"
              using \<open>vars_term r\<^sub>2 \<subseteq> vars_term l\<^sub>2 \<union> vars_trs (set cs\<^sub>2)\<close> snd_conv by auto
          qed
        then have "\<forall>x \<in> vars_crule ?\<rho>\<^sub>2. (\<sigma>\<^sub>1' x, \<sigma>\<^sub>2' x) \<in> (cstep R)\<^sup>\<down>"
          using evars_join substl \<sigma>\<^sub>1' \<sigma>\<^sub>2' \<sigma>\<^sub>1'_def \<sigma>\<^sub>2'_def by auto
        then have "\<forall>x \<in> vars_crule ?\<rho>\<^sub>2. (?\<sigma>\<^sub>1 x, \<sigma>\<^sub>2 x) \<in> (cstep R)\<^sup>\<down>"
          by (meson \<sigma>\<^sub>1' \<sigma>\<^sub>2' join_rtrancl_join rtrancl_join_join)
        then show ?thesis unfolding t1 t2
          by (metis Un_iff fst_conv prod.sel(2) term_subst_csteps_join vars_crule_def vars_defs(2))
      next
        case (proper)
        def \<sigma>' \<equiv> "\<lambda>x. if x \<in> vars_crule ?\<rho> then \<sigma> x else \<sigma>\<^sub>2 x"
        have unif: "l\<^sub>2 \<cdot> \<sigma>' = (l |_ p) \<cdot> \<sigma>'"
        proof -
          note coinc = coincidence_lemma' [of l\<^sub>2 "vars_crule ?\<rho>\<^sub>2"]
          have disj: "vars_crule ?\<rho>\<^sub>2 \<inter> vars_term (l |_ p) = {}"
            using vars_term_subt_at [OF poss] and disj
            by (auto simp: vars_crule_def vars_rule_def)
          have "l\<^sub>2 \<cdot> \<sigma>' = l\<^sub>2 \<cdot> (\<sigma>' |s vars_crule ?\<rho>\<^sub>2)"
            using coinc by (simp add: vars_crule_def vars_rule_def sup_assoc)
          also have "\<dots> = l\<^sub>2 \<cdot> (\<sigma>\<^sub>2 |s vars_crule ?\<rho>\<^sub>2)" apply (simp add: \<sigma>'_def)
            by (metis (no_types) \<pi> crule_pt.permute_prod.simps cs_def l_def r_def
                rule_pt.permute_prod.simps subst_restrict_restrict)
          also have "\<dots> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using coinc by (simp add: vars_crule_def vars_rule_def) blast
          also have "\<dots> = (l \<cdot> \<sigma>) |_ p" by (simp add: \<open>l |_ p \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> poss)
          also have "\<dots> = (l |_ p) \<cdot> \<sigma>" by (simp add: \<open>l |_ p \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> \<open>l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = l \<cdot> \<sigma> |_ p\<close>) 
          also have "\<dots> = (l |_ p) \<cdot> (\<sigma> |s vars_term (l |_ p))"
            by (simp add: coincidence_lemma [symmetric])
          also have "\<dots> = (l |_ p) \<cdot> (\<sigma>' |s vars_term (l |_ p))" using disj
            apply (simp add: \<sigma>'_def)
            by (metis (no_types) \<open>vars_term (l |_ p) \<subseteq> vars_term l\<close> coincidence_lemma coincidence_lemma' fst_conv subst_restrict_vars sup.coboundedI1 vars_crule_def vars_rule_def)
          finally show ?thesis by (simp add: coincidence_lemma [symmetric])
        qed
        from is_imgu_imp_is_mgu [OF mgu_sound [OF mgu]] have "is_mgu \<tau> {(l |_ p, l\<^sub>2)}" .
        with unif obtain \<delta> where \<sigma>': "\<sigma>' = \<tau> \<circ>\<^sub>s \<delta>"
          by (auto simp: is_mgu_def unifiers_def) presburger
        from proper CCP_I [OF rule_variants disj funposs] and mgu
          have ccp: "((r \<cdot> \<tau>, (ctxt_of_pos_term p l)\<langle>r\<^sub>2\<rangle> \<cdot> \<tau>), subst_list \<tau> (cs @ cs\<^sub>2)) \<in> CCP R"
          by simp
        have sig:  "\<forall>(s, t) \<in> set cs. \<forall>x \<in> vars_term s \<union> vars_term t. \<sigma> x = \<sigma>' x"
          by (fastforce simp: \<sigma>'_def vars_crule_def vars_trs_def vars_rule_def)
        { fix s t
          assume a: "(s, t) \<in> set cs"
          with sig have "\<forall>x \<in> vars_term s. \<sigma> x = \<sigma>' x" and "\<forall>x \<in> vars_term t. \<sigma> x = \<sigma>' x"
            by auto
          moreover with cs a have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" by fast
          ultimately have "(s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*"
            using term_subst_eq by metis
        }
        then have 1: "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*" by auto
        have sig2:  "\<forall>(s, t) \<in> set cs\<^sub>2. \<forall>x \<in> vars_term s \<union> vars_term t. \<sigma>\<^sub>2 x = \<sigma>' x"
          using disj by (fastforce simp: \<sigma>'_def vars_crule_def vars_trs_def vars_rule_def)
        { fix s t
          assume a: "(s, t) \<in> set cs\<^sub>2"
          with sig2 have "\<forall>x \<in> vars_term s. \<sigma>\<^sub>2 x = \<sigma>' x" and "\<forall>x \<in> vars_term t. \<sigma>\<^sub>2 x = \<sigma>' x"
            by auto
          moreover with cs2 a have "(s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*" by fast
          ultimately have "(s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*"
            using term_subst_eq by metis
        }
        then have 2: "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*" by auto

        have t1: "t\<^sub>1 = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" by (metis C C' ctxt_apply_term.simps(1) ctxt_supt
            subterm.dual_order.strict_implies_not_eq t)
        from t tp' have "t = (ctxt_of_pos_term p (l\<^sub>1 \<cdot> \<sigma>\<^sub>1))\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
          by (metis \<alpha> ctxt_supt_id funposs_imp_poss poss_imp_subst_poss)
        then have **: "D = ctxt_of_pos_term p (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using p D
          using ctxt_of_pos_term_hole_pos t by force
        then have t2: "t\<^sub>2 = (ctxt_of_pos_term p (l\<^sub>1 \<cdot> \<sigma>\<^sub>1))\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
          using D' [unfolded **] by simp
        from rename have "\<sigma>\<^sub>1 = sop \<pi> \<circ>\<^sub>s \<sigma>" unfolding subst_compose_def o_def
          by (metis inv_Rep_perm_simp permute_atom_def subst_apply_term.simps(1))
        from t1 rename have "t\<^sub>1 = (-\<pi>) \<bullet> r \<cdot> (sop \<pi> \<circ>\<^sub>s \<sigma>)" by force
        then have "t\<^sub>1 = r \<cdot> \<sigma>" by simp
        then have t1': "t\<^sub>1 = r \<cdot> \<sigma>'" using \<sigma>'_def
          by (simp add: UnCI term_subst_eq_conv vars_crule_def vars_defs(2))
        from t2 rename have "t\<^sub>2 = (ctxt_of_pos_term p (l \<cdot> \<sigma>))\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>" by force
        moreover have "l \<cdot> \<sigma> = l \<cdot> \<sigma>'" using \<sigma>'_def
          by (simp add: UnCI term_subst_eq_conv vars_crule_def vars_defs(2))
        moreover have "r\<^sub>2 \<cdot> \<sigma>\<^sub>2 = r\<^sub>2 \<cdot> \<sigma>'" using disj
          by (induct r\<^sub>2) (auto simp: \<sigma>'_def vars_crule_def vars_rule_def)
        ultimately have "t\<^sub>2 = (ctxt_of_pos_term p (l \<cdot> \<sigma>'))\<langle>r\<^sub>2 \<cdot> \<sigma>'\<rangle>" by presburger
        then have t2': "t\<^sub>2 = (ctxt_of_pos_term p l)\<langle>r\<^sub>2\<rangle> \<cdot> \<sigma>'"
          by (simp add: ctxt_of_pos_term_subst poss)

        from 1 2 have "\<forall>(s, t) \<in> set (cs @ cs\<^sub>2). (s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*"
          using UnE set_append by auto
        then obtain n where "conds_sat R n (cs @ cs\<^sub>2) \<sigma>'" using conds_sat_iff' by blast
        from conds_sat_subst_list [THEN iffD2, OF this [unfolded \<sigma>']]
          have "conds_sat R n (subst_list \<tau> (cs @ cs\<^sub>2)) \<delta>" .
        (* difference to 4.1 start *)
        with ccp have "(r \<cdot> \<tau> \<cdot> \<delta>, (ctxt_of_pos_term p l)\<langle>r\<^sub>2\<rangle> \<cdot> \<tau> \<cdot> \<delta>) \<in> (cstep R)\<^sup>\<down>"
          using allCPs [unfolded all_CPs_joinable_def] by blast
        then have join: "(r \<cdot> \<sigma>', (ctxt_of_pos_term p l)\<langle>r\<^sub>2\<rangle> \<cdot> \<sigma>') \<in> (cstep R)\<^sup>\<down>" using \<sigma>' by simp
        from join t1' t2' show ?thesis by simp
        (* difference to 4.2 end *)
      qed
    next
      case (\<beta>)
      moreover have \<beta>':  "p \<in> poss (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using D p t by auto
      from poss_subst_apply_term [OF \<beta>' \<beta>]
        obtain q\<^sub>1 q\<^sub>2 x where [simp]: "p = q\<^sub>1 <#> q\<^sub>2" and q\<^sub>1: "q\<^sub>1 \<in> poss l\<^sub>1"
      and l\<^sub>1q\<^sub>1: "l\<^sub>1 |_ q\<^sub>1 = Var x" and q\<^sub>2: "q\<^sub>2 \<in> poss (\<sigma>\<^sub>1 x)" by auto
      moreover
        have [simp]: "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 |_ p = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using tp' [unfolded t] .
      ultimately
        have [simp]: "\<sigma>\<^sub>1 x |_ q\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" by simp
  
      have \<sigma>1step: "(\<sigma>\<^sub>1 x, replace_at (\<sigma>\<^sub>1 x) q\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) \<in> cstep R" (is "(_, ?t\<^sub>0) \<in> _")
        using cstep_ctxt s2 by (metis \<open>\<sigma>\<^sub>1 x |_ q\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> ctxt_supt_id q\<^sub>2)
      
      def \<tau> \<equiv> "\<lambda>y. if y = x then ?t\<^sub>0 else \<sigma>\<^sub>1 y"

      obtain \<tau>' where norm_\<tau>': "normalized R \<tau>'" and \<tau>': "\<forall>x. (\<tau> x, \<tau>' x) \<in> (cstep R)\<^sup>*"
        using quasi_decreasing_obtains_normalized_subst [OF qro] by auto
      
      have "(\<sigma>\<^sub>1 x, \<tau> x) \<in> cstep R"
      proof -
        let ?C = "ctxt_of_pos_term q\<^sub>2 (\<sigma>\<^sub>1 x)"
        have "(?C\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>, ?C\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>) \<in> cstep R" using s2 by (simp add: cstep_ctxt) 
        then show ?thesis using q\<^sub>2 by (simp add: \<tau>_def replace_at_ident)
      qed
      then have *: "\<And>x. (\<sigma>\<^sub>1 x, \<tau> x) \<in> (cstep R)\<^sup>*" by (auto simp: \<tau>_def)
  
      have "(t\<^sub>2, l\<^sub>1 \<cdot> \<tau>) \<in> (cstep R)\<^sup>*"
      proof -
        have "replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2) = replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) q\<^sub>1 (\<tau> x)"
          using q\<^sub>1 and q\<^sub>2 by (simp add: \<tau>_def ctxt_of_pos_term_append l\<^sub>1q\<^sub>1)
        moreover
        have "(replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) q\<^sub>1 (\<tau> x), l\<^sub>1 \<cdot> \<tau>) \<in> (cstep R)\<^sup>*"
          by (rule replace_at_subst_csteps [OF * q\<^sub>1 l\<^sub>1q\<^sub>1])
        then show ?thesis
          by (metis D D' calculation ctxt_of_pos_term_hole_pos p t)
      qed
      then have t2csteps: "(t\<^sub>2, l\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" unfolding t using \<tau>'
        by (meson * rtrancl_trans substs_csteps)
      
      have "\<forall>x \<in> vars_term r\<^sub>1. (\<sigma>\<^sub>1 x, \<tau> x) \<in> (cstep R)\<^sup>*" using * \<sigma>1step rtrancl.simps by auto
      then have "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<tau>) \<in> (cstep R)\<^sup>*" using term_subst_csteps [of r\<^sub>1 \<sigma>\<^sub>1 \<tau>] by blast
      moreover have "(r\<^sub>1 \<cdot> \<tau>, r\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" using \<tau>' by (simp add: substs_csteps)
      ultimately have "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by auto
      then have t1csteps: "(t\<^sub>1, r\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" using t s1
        by (metis C C' ctxt_apply_term.simps(1) ctxt_supt supt_neqD)
      
      { fix u v
        assume a: "(u, v) \<in> set cs\<^sub>1"
        have cs1': "\<forall>(s, t) \<in> set cs\<^sub>1. (s \<cdot> \<sigma>\<^sub>1, t \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
          using cs1 by (auto simp: csteps_n_imp_csteps)
        from a have "(u \<cdot> \<sigma>\<^sub>1, v \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*" using cs1 by (auto simp: csteps_n_imp_csteps)
        then have 1: "(u \<cdot> \<sigma>\<^sub>1, v \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by (meson * \<tau>' rtrancl_trans substs_csteps)
        
        from quasi_decreasing_order_conds_sat [OF qro r1 cs1']
          have 2: "\<forall>i < length cs\<^sub>1. (fst (cs\<^sub>1 ! i) \<cdot> \<sigma>\<^sub>1, l\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" by auto
        have 3: "(u \<cdot> \<sigma>\<^sub>1, u \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by (meson * \<tau>' rtrancl_trans substs_csteps)
        have "(u \<cdot> \<sigma>\<^sub>1, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" unfolding t using a 2 by (metis fst_conv in_set_idx)
        from IH [OF this 3 1] have 4: "(u \<cdot> \<tau>', v \<cdot> \<tau>') \<in> (cstep R)\<^sup>\<down>" .
        
        from SDTRS [unfolded sdtrs_def strongly_deterministic_def] r1
            have "\<forall>(u, v) \<in> set cs\<^sub>1. strongly_irreducible R v" by fast
        then have siv: "strongly_irreducible R v" using a  by auto
        from norm_\<tau>' siv have "v \<cdot> \<tau>' \<in> NF (cstep R)" by (auto simp: strongly_irreducible_def)
        with 4 have "(u \<cdot> \<tau>', v \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by (simp add: NF_join_imp_reach)
      }
      then have "\<forall>(u, v) \<in> set cs\<^sub>1. (u \<cdot> \<tau>', v \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by auto
      from cstepI [OF r1 this, of "l\<^sub>1 \<cdot> \<tau>'" \<box> "r\<^sub>1 \<cdot> \<tau>'"]
        have "(l\<^sub>1 \<cdot> \<tau>', r\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by simp
      with t1csteps t2csteps
       show ?thesis using joinI rtrancl_trans by auto
    qed
  qed
qed

text \<open>Theorem 4.1 from AL94, but using quasi-decreasingness\<close>
theorem quasi_decreasing_sdtrs_all_CPs_joinable_CR:
  assumes sdtrs: "sdtrs R"
    and qd: "quasi_decreasing R"
    and allCPs: "all_CPs_joinable R"
  shows "CR (cstep R)"
proof -
  { fix t t' t''
    from qd obtain S where qdo: "quasi_decreasing_order R S" by (auto simp: quasi_decreasing_def)
    then have SNS: "SN S" by (auto simp: quasi_decreasing_order_def)
    then have "SN ((S \<union> {\<rhd>})\<^sup>+)" (is "SN ?S")
    proof -
      have "{\<rhd>} \<subseteq> S"
        using qdo quasi_decreasing_order_def by blast
      then show ?thesis
        by (simp add: SNS SN_imp_SN_trancl subset_Un_eq sup_commute)
    qed
    from qdo have trans: "trans S" and RsubS: "cstep R \<subseteq> S"
      by (auto simp: quasi_decreasing_order_def)
    then have *: "cstep R \<subseteq> ?S" using qdo quasi_decreasing_order_def by auto

    assume 1: "(t, t') \<in> (cstep R)\<^sup>*" and 2: "(t, t'') \<in> (cstep R)\<^sup>*"
    then have "(t', t'') \<in> (cstep R)\<^sup>\<down>"
    proof (induct t arbitrary: t' t'' rule: wf_induct[OF SN_imp_wf[OF \<open>SN ?S\<close>], rule_format])
      case (1 t)
      note IH = this(1)
      { assume "t = t' \<or> t = t''"
        then have "(t', t'') \<in> (cstep R)\<^sup>\<down>" using 1 2 by blast
      }
      moreover
      { assume "\<exists>t\<^sub>1. (t, t\<^sub>1) \<in> cstep R \<and> (t\<^sub>1, t') \<in> (cstep R)\<^sup>*"
          and "\<exists>t\<^sub>2. (t, t\<^sub>2) \<in> cstep R \<and> (t\<^sub>2, t'') \<in> (cstep R)\<^sup>*"
        then obtain t\<^sub>1 t\<^sub>2 where t11: "(t, t\<^sub>1) \<in> cstep R" and t12:"(t\<^sub>1, t') \<in> (cstep R)\<^sup>*"
          and t21: "(t, t\<^sub>2) \<in> cstep R" and t22: "(t\<^sub>2, t'') \<in> (cstep R)\<^sup>*" by blast
        from t11 obtain n' where "(t, t\<^sub>1) \<in> cstep_n R n'" using cstep_iff by blast 
        from cstep_nE [OF this] obtain C l\<^sub>1 r\<^sub>1 \<sigma>\<^sub>1 cs\<^sub>1 n
          where r1: "((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R"
          and "n' = Suc n"
          and cs1: "\<forall>(s, t) \<in> set cs\<^sub>1. (s \<cdot> \<sigma>\<^sub>1, t \<cdot> \<sigma>\<^sub>1) \<in> (cstep_n R n)\<^sup>*"
          and C: "t = C\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>"
          and C': "t\<^sub>1 = C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>" by metis
        from t21 obtain m' where "(t, t\<^sub>2) \<in> cstep_n R m'" using cstep_iff by blast
        from cstep_nE [OF this] obtain D l\<^sub>2 r\<^sub>2 \<sigma>\<^sub>2 cs\<^sub>2 m
          where r2: "((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R"
          and "m' = Suc m"
          and cs2: "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep_n R m)\<^sup>*"
          and D: "t = D\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
          and D': "t\<^sub>2 = D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>" by metis
        from cstep_n_SucI [OF r1 cs1, of "l\<^sub>1 \<cdot> \<sigma>\<^sub>1" \<box> "r\<^sub>1 \<cdot> \<sigma>\<^sub>1"]
          have s1: "(l\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> cstep R" by (simp add: cstep_n_imp_cstep)
        from cstep_n_SucI [OF r2 cs2, of "l\<^sub>2 \<cdot> \<sigma>\<^sub>2" \<box> "r\<^sub>2 \<cdot> \<sigma>\<^sub>2"]
          have s2: "(l\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> cstep R" by (simp add: cstep_n_imp_cstep)
        obtain q p where q: "hole_pos C = q" and p: "hole_pos D = p" by simp
        have tq': "t |_q = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using C q by fastforce
        from cs1 have cs1': "\<forall>(s, t) \<in> set cs\<^sub>1. (s \<cdot> \<sigma>\<^sub>1, t \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
          using csteps_n_imp_csteps by blast
        from cs2 have cs2': "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
          using csteps_n_imp_csteps by blast
        consider (GE) "q \<le> p" | (LE) "p \<le> q" | (PL) "q \<bottom> p" using parallel_pos by blast 
        then have "(t\<^sub>1, t\<^sub>2) \<in> (cstep R)\<^sup>\<down>"
        proof (cases)
          case (GE)
          from aux_lemma [OF qdo allCPs sdtrs r1 cs1' C C' r2 cs2' D D' q p GE IH]
            show ?thesis .
        next
          case (LE)
          from aux_lemma [OF qdo allCPs sdtrs r2 cs2' D D' r1 cs1' C C' p q LE IH]
            have "(t\<^sub>2, t\<^sub>1) \<in> (cstep R)\<^sup>\<down>" .
          then show ?thesis by auto
        next
          case (PL)
          from C D q p have tpar: "replace_at t q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) = replace_at t p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)"
            using hole_pos_id_ctxt by fastforce
          have *: "t = replace_at (replace_at t q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" using C tpar q by auto
          have **: "t = replace_at (replace_at t p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using C tpar q by auto
          then have "C\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle> = replace_at (replace_at t q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" using * C by blast
          then have "C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle> = replace_at (replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)"
            by (metis C D PL ctxt_of_pos_term_hole_pos hole_pos_poss p parallel_replace_at q)
          then have t1: "t\<^sub>1 = replace_at (replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" using C' by blast
          have 1: "(t\<^sub>1, replace_at (replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) \<in> cstep R"
            (is "(_, ?t4) \<in> _") unfolding t1 using s2 by (simp add: cstep_ctxt)
          from ** have "D\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle> = replace_at (replace_at t p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using D by blast
          then have "D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle> = replace_at (replace_at t p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)"
            using D p tpar by (metis C PL ctxt_of_pos_term_hole_pos hole_pos_poss parallel_replace_at q)
          then have t2: "t\<^sub>2 = replace_at (replace_at t p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using D' by blast
          have 2: "(t\<^sub>2, replace_at (replace_at t p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)) \<in> cstep R"
            (is "(_, ?t3) \<in> _") unfolding t2 using s1 by (simp add: cstep_ctxt)
          have "?t3 = ?t4" using PL by (metis C D hole_pos_poss p parallel_replace_at q)
          then show ?thesis using 1 2 by auto
        qed
        then obtain s where s01: "(t\<^sub>1, s) \<in> (cstep R)\<^sup>*" and s02: "(t\<^sub>2, s) \<in> (cstep R)\<^sup>*" by blast
        from t11 * have "(t\<^sub>1, t) \<in> ?S\<inverse>" by blast
        from IH [OF this s01 t12] have "(s, t') \<in> (cstep R)\<^sup>\<down>" .
        then obtain s' where s11: "(t', s') \<in> (cstep R)\<^sup>*" and s12: "(s, s') \<in> (cstep R)\<^sup>*" by blast
        from t21 * have t2tqr: "(t\<^sub>2, t) \<in> ?S\<inverse>" by blast
        have "(t\<^sub>2, s') \<in> (cstep R)\<^sup>*" using s02 s12 by force
        from IH [OF t2tqr this t22] have "(s', t'') \<in> (cstep R)\<^sup>\<down>" .
        with s11 have "(t', t'') \<in> (cstep R)\<^sup>\<down>" by (simp add: rtrancl_join_join)
      }
      ultimately show "(t', t'') \<in> (cstep R)\<^sup>\<down>" using 1 2 converse_rtranclE by metis
    qed
  }
  then show "CR (cstep R)" unfolding CR_on_def by force
qed

text \<open>Theorem 4.1 from AL94\<close>
theorem quasi_reductive_sdtrs_all_CPs_joinable_CR:
  assumes sdtrs: "sdtrs R"
    and qd: "quasi_reductive R"
    and allCPs: "all_CPs_joinable R"
  shows "CR (cstep R)"
using assms
by (blast dest: quasi_reductive_quasi_decreasing quasi_decreasing_sdtrs_all_CPs_joinable_CR)

definition absolutely_irreducible :: "('f, 'v :: infinite) ctrs \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "absolutely_irreducible R v \<longleftrightarrow> (\<forall>\<rho> \<in> R. \<forall>p \<in> funposs v. \<forall>\<pi>. \<not> unifiable {(v |_ p, \<pi> \<bullet> clhs \<rho>)})"

definition absolutely_deterministic :: "('f, 'v :: infinite) ctrs \<Rightarrow> bool"
where
  "absolutely_deterministic R \<longleftrightarrow> (\<forall>\<rho> \<in> R. \<forall>(u, v) \<in> set (conds \<rho>). absolutely_irreducible R v)"

(* TODO: move (?) *)
lemma poss_is_Fun_funposs:
  assumes "p \<in> poss t"
    and "is_Fun (t |_ p)"
  shows "p \<in> funposs t"
using assms by (metis DiffI is_Var_def poss_simps(3) varposs_iff)

(* TODO: move (?) *)
lemma disj_vars_term_matchers_unifieres:
  assumes disj: "vars_term s \<inter> vars_term t = {}"
    and matcher: "\<sigma> \<in> matchers {(s, t)}"
  shows "(\<sigma> |s vars_term s) \<in> unifiers {(t, s)}" (is "?\<sigma> \<in> _")
proof -
  from matcher have "s \<cdot> \<sigma> = t" unfolding matchers_def by force
  moreover have "s \<cdot> \<sigma> = s \<cdot> (\<sigma> |s vars_term s)" using coincidence_lemma by blast
  ultimately have "s \<cdot> ?\<sigma> = t \<cdot> ?\<sigma>" using disj by (simp add: inf_commute subst_apply_id')
  then show ?thesis unfolding unifiers_def by auto
qed

text \<open>Lemma 4.1a from AL94\<close>
lemma absolutely_irreducible_strongly_irreducible:
  assumes "absolutely_irreducible R v"
    and vc: "\<forall>((l, r), cs) \<in> R. is_Fun l"
  shows "strongly_irreducible R v"
proof (rule ccontr)
  assume "\<not> strongly_irreducible R v"
  then obtain \<tau> where norm: "normalized R \<tau>" and "\<not> v \<cdot> \<tau> \<in> NF (cstep R)" (is "\<not> ?v \<in> _")
    unfolding strongly_irreducible_def by blast
  then obtain v' where "(?v, v') \<in> (cstep R)" by blast
  from cstepE[OF this] obtain C \<sigma> l r cs where rule: "((l, r), cs) \<in> R"
    and cs: "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
    and v: "?v = C\<langle>l \<cdot> \<sigma>\<rangle>" and "v' = C\<langle>r \<cdot> \<sigma>\<rangle>" by metis
  then obtain p where "hole_pos C = p" and p: "p \<in> poss ?v" by auto
  with v norm have match: "?v |_ p = l \<cdot> \<sigma>" unfolding normalized_def by fastforce
  with p have "is_Fun (?v |_ p)" using rule vc by (cases l) auto
  then have p': "p \<in> funposs ?v" using poss_is_Fun_funposs p by blast
  have pfv: "p \<in> funposs v"
  proof (rule ccontr)
    assume "p \<notin> funposs v"
    with p' [THEN funposs_imp_poss] obtain q x where "\<tau> x |_ q = l \<cdot> \<sigma>" and "q \<in> poss (\<tau> x)"
      using match by (auto elim!: poss_subst_apply_term)
    with cstepI [OF rule cs] have "(\<tau> x, (ctxt_of_pos_term q (\<tau> x))\<langle>r \<cdot> \<sigma>\<rangle>) \<in> (cstep R)"
      by auto (metis replace_at_ident)
    then show "False" using norm by (auto simp: normalized_def)
  qed
  then have pv: "p \<in> poss v" using funposs_imp_poss by auto
  have "\<exists>\<pi>. vars_term (\<pi> \<bullet> l) \<inter> (vars_term (?v |_ p) \<union> vars_term (v |_ p)) = {}"
  proof -
    obtain \<pi> where "term_pt.supp (\<pi> \<bullet> l) \<sharp> (?v |_ p, v |_ p)"
      by (metis finite_term_supp rule_fs.atom_set_avoiding supp_vars_term_eq vars_term_eqvt)
    from rule_pt.fresh_set_disjoint [OF this]
      show ?thesis by (auto simp: supp_vars_rule_eq vars_rule_def supp_vars_term_eq)
  qed
  then obtain \<pi> where disj: "vars_term (\<pi> \<bullet> l) \<inter> vars_term (?v |_ p) = {}"
    and disj': "vars_term (\<pi> \<bullet> l) \<inter> vars_term (v |_ p) = {}" by blast
  from match have \<sigma>: "\<sigma> \<in> matchers {(l, ?v |_ p)}" unfolding matchers_def by auto
  then have "sop (-\<pi>) \<circ>\<^sub>s \<sigma> \<in> matchers {(\<pi> \<bullet> l, ?v |_ p)}" (is "?\<sigma> \<in> _") unfolding matchers_def by auto
  from disj_vars_term_matchers_unifieres [OF disj this]
    have "?\<sigma> |s vars_term (\<pi> \<bullet> l) \<in> unifiers {(?v |_ p, \<pi> \<bullet> l)}" (is "?\<sigma>' \<in> _") .
  then have "?v |_ p \<cdot> ?\<sigma>' = \<pi> \<bullet> l \<cdot> ?\<sigma>'" unfolding unifiers_def by fastforce
  moreover have "?v |_ p \<cdot> ?\<sigma>' = v |_ p \<cdot> (\<tau> |s vars_term (v |_ p)) \<circ>\<^sub>s ?\<sigma>'" (is "_ = _ \<cdot> ?\<tau> \<circ>\<^sub>s _")
    using pv by (metis (no_types) coincidence_lemma subst_subst subt_at_subst)
  moreover have "\<pi> \<bullet> l \<cdot> ?\<sigma>' = \<pi> \<bullet> l \<cdot> ?\<tau> \<circ>\<^sub>s ?\<sigma>'" (is "_ = _ \<cdot> ?\<tau>'")
    using disj' by (simp add: subst_apply_id')
  ultimately have "v |_ p \<cdot> ?\<tau>' = \<pi> \<bullet> l \<cdot> ?\<tau>'" by auto
  then have "?\<tau>' \<in> unifiers {(v |_ p, \<pi> \<bullet> l)}" unfolding unifiers_def by auto
  then have "unifiable {(v |_ p, \<pi> \<bullet> l)}" unfolding unifiable_def by auto
  moreover from assms(1) [unfolded absolutely_irreducible_def] have "\<not> unifiable {(v |_ p, \<pi> \<bullet> l)}"
    using rule pfv fst_conv by auto
  ultimately show "False" by auto
qed

text \<open>Lemma 4.1b from AL94\<close>
lemma absolutely_deterministic_strongly_deterministic:
  assumes "absolutely_deterministic R"
    and vc: "\<forall>((l, r), cs) \<in> R. is_Fun l"
  shows "strongly_deterministic R"
proof -
  { fix \<rho>
    assume "\<rho> \<in> R"
    with assms have *: "\<forall>(u, v) \<in> set (conds \<rho>). absolutely_irreducible R v"
      by (auto simp: absolutely_deterministic_def)
    { fix u v
      assume "(u, v) \<in> set (conds \<rho>)"
      then have "absolutely_irreducible R v" using * by fast
      from absolutely_irreducible_strongly_irreducible [OF this vc]
        have "strongly_irreducible R v" .
    }
    then have "\<forall>(u, v) \<in> set (conds \<rho>). strongly_irreducible R v" by auto
  }
  then have "\<forall> ((l, r), cs) \<in> R. \<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs. strongly_irreducible R t\<^sub>i"
    by (simp add: case_prodI2)
  then show ?thesis by (auto simp: strongly_deterministic_def)
qed

text \<open>Definition 4.4 from AL94\<close>
definition "unfeasible R S r r' p \<longleftrightarrow>
  dctrs R \<and>
  overlap R r r' p \<and>
  (let \<sigma> = the (mgu (clhs r |_ p) (clhs r')); C = set (subst_list \<sigma> (snd r @ snd r')) in
    \<exists> t\<^sub>0 t\<^sub>1 t\<^sub>2. (clhs r \<cdot> \<sigma>, t\<^sub>0) \<in> S \<and>
                (t\<^sub>0, t\<^sub>1) \<in> (ctxt_step (cstep R) C)\<^sup>* \<and>
                (t\<^sub>0, t\<^sub>2) \<in> (ctxt_step (cstep R) C)\<^sup>* \<and>
                \<not> unifiable {(t\<^sub>1, t\<^sub>2)} \<and>
                strongly_irreducible R t\<^sub>1 \<and>
                strongly_irreducible R t\<^sub>2
  )"

text \<open>Definition 4.4 from AL94\<close>
definition "context_joinable R r r' p \<longleftrightarrow>
  dctrs R \<and>
  (let \<sigma> = the (mgu (clhs r |_ p) (clhs r'));
    s = crhs r \<cdot> \<sigma>; t = replace_at (clhs r) p (crhs r') \<cdot> \<sigma>;
    C = set (subst_list \<sigma> (snd r @ snd r')) in
    \<exists> t\<^sub>0. (s, t\<^sub>0) \<in> (ctxt_step (cstep R) C)\<^sup>* \<and> (t, t\<^sub>0) \<in> (ctxt_step (cstep R) C)\<^sup>*
  )"

lemma unfeasible_supt_trancl:
  assumes "unfeasible R S r r' p"
  shows "unfeasible R ((S \<union> {\<rhd>})\<^sup>+) r r' p"
using assms by (auto simp: unfeasible_def Let_def)

lemma aux_lemma':
  fixes l\<^sub>1 :: "('f, 'v :: infinite) term"
  assumes S: "quasi_decreasing_order R S"
    and subst: "subst.closed S"
    and allCPs: "\<forall> r r' p. overlap R r r' p \<longrightarrow> unfeasible R S r r' p \<or> context_joinable R r r' p"
    and SDTRS: "sdtrs R"
    and r1: "((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R" (is "?\<rho>\<^sub>1 \<in> _")
    and cs1: "\<forall>(s, t) \<in> set cs\<^sub>1. (s \<cdot> \<sigma>\<^sub>1, t \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
    and C: "t = C\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>"
    and C': "t\<^sub>1 = C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>"
    and r2: "((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R" (is "?\<rho>\<^sub>2 \<in> _")
    and cs2: "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
    and D: "t = D\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
    and D': "t\<^sub>2 = D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
    and q: "hole_pos C = q"
    and p: "hole_pos D = p"
    and GE: "q \<le> p"
    and IH: "\<And>y t' t''. (y, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse> \<Longrightarrow> (y, t') \<in> (cstep R)\<^sup>*
      \<Longrightarrow> (y, t'') \<in> (cstep R)\<^sup>* \<Longrightarrow> (t', t'') \<in> (cstep R)\<^sup>\<down>"
  shows "(t\<^sub>1, t\<^sub>2) \<in> (cstep R)\<^sup>\<down>"
proof -
  have tq': "t |_q = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using C q by fastforce
  from cstepI [OF r1 cs1, of "l\<^sub>1 \<cdot> \<sigma>\<^sub>1" \<box> "r\<^sub>1 \<cdot> \<sigma>\<^sub>1"] have s1: "(l\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> cstep R" by simp
  from cstepI [OF r2 cs2, of "l\<^sub>2 \<cdot> \<sigma>\<^sub>2" \<box> "r\<^sub>2 \<cdot> \<sigma>\<^sub>2"] have s2: "(l\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> cstep R" by simp
  from GE have r: "p = q <#> pos_diff p q" (is "_ = _ <#> ?r") by simp
  consider (ST) "t \<rhd> t |_q" | (EQ) "t = t |_q"
    using C hole_pos_poss q subt_at_imp_supteq by blast
  then show ?thesis
  proof (cases)
    case (ST)
    from GE ST have st: "t |_q \<unrhd> t |_p" using D hole_pos_poss less_eq_pos_imp_supt_eq p by blast 
    have *: "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 = replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) ?r (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)"
    proof -
      have "hole_pos (D |_c q) = pos_diff p q" using r hole_pos_subt_at_ctxt p by blast
      then show ?thesis by (metis D r tq' ctxt_of_pos_term_hole_pos p subt_at_subt_at_ctxt)
    qed
    have tq: "(t |_q, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" using ST by blast
    have 1: "(t |_q, r\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*" unfolding tq' using s1 by blast
    have 2: "(t |_q, replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) ?r (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) \<in> (cstep R)\<^sup>*" (is "(_, ?t2) \<in> _")
      unfolding tq' by (metis * csteps_ctxt r_into_rtrancl s2)
    from IH [OF tq 1 2] have j: "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, ?t2) \<in> (cstep R)\<^sup>\<down>" .
    then obtain s where "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, s) \<in> (cstep R)\<^sup>*" and "(?t2, s) \<in> (cstep R)\<^sup>*" by blast
    then have 5: "(replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1), replace_at t q s) \<in> (cstep R)\<^sup>*"
      and 6: "(replace_at t q ?t2, replace_at t q s) \<in> (cstep R)\<^sup>*"
      by (simp add: csteps_ctxt)+
    have 3: "t\<^sub>1 = replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using C C' q by auto
    have 4: "t\<^sub>2 = replace_at t q ?t2" using r
      by (metis C D D' ctxt_ctxt ctxt_of_pos_term_append
          ctxt_of_pos_term_hole_pos hole_pos_poss p q tq')
    show ?thesis using 5 6 unfolding 3 4 by blast
  next
    case (EQ)
    have t: "t = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using EQ tq' by simp
    have tp': "t |_ p = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using D p by fastforce
    consider (\<alpha>) "p \<in> funposs l\<^sub>1" | (\<beta>) "p \<notin> funposs l\<^sub>1" by auto
    then show ?thesis
    proof (cases)
      case (\<alpha>)
      from vars_crule_disjoint obtain \<pi>
        where \<pi>: "vars_crule (\<pi> \<bullet> ?\<rho>\<^sub>1) \<inter> vars_crule ?\<rho>\<^sub>2 = {}" ..
      def l \<equiv> "\<pi> \<bullet> l\<^sub>1" and r \<equiv> "\<pi> \<bullet> r\<^sub>1" and cs \<equiv> "\<pi> \<bullet> cs\<^sub>1" and \<sigma> \<equiv> "sop (-\<pi>) \<circ>\<^sub>s \<sigma>\<^sub>1"
      note rename = l_def r_def cs_def \<sigma>_def
      
      have *: "-\<pi> \<bullet> ((l, r), cs) \<in> R" (is "-\<pi> \<bullet> ?\<rho> \<in> R") and "0 \<bullet> ?\<rho>\<^sub>2 \<in> R"
        using r1 r2 by (simp_all add: eqvt rename o_def)
      then have rule_variants: "\<exists>\<pi>. \<pi> \<bullet> ?\<rho> \<in> R" "\<exists>\<pi>. \<pi> \<bullet> ?\<rho>\<^sub>2 \<in> R"
        by blast+
  
      have disj: "vars_crule ?\<rho> \<inter> vars_crule ?\<rho>\<^sub>2 = {}"
        using \<pi> by (auto simp: eqvt rename)
      
      from cs1 have cs: "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
        by (auto simp: rename rule_pt.permute_prod.simps)
        
      from \<alpha> have funposs: "p \<in> funposs (clhs ?\<rho>)" by (simp add: rename)
      then have poss: "p \<in> poss l" by (auto dest: funposs_imp_poss)
      with vars_term_subt_at [OF this] and disj
        have "vars_term (l |_ p) \<inter> vars_term l\<^sub>2 = {}" and "(l |_ p) \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
        apply (auto simp: eqvt vars_crule_def vars_rule_def rename)
        using EQ tp' tq' by auto
      from vars_term_disjoint_imp_unifier [OF this]
        obtain \<tau> where mgu: "mgu (l |_ p) l\<^sub>2 = Some \<tau>"
        using mgu_complete by (auto simp: unifiers_def)
      consider (improp) "p = \<epsilon> \<and> (\<exists>\<pi>. \<pi> \<bullet> ?\<rho> = ?\<rho>\<^sub>2)"
        | (proper) "\<not>p = \<epsilon> \<or> \<not>(\<exists>\<pi>. \<pi> \<bullet> ?\<rho> = ?\<rho>\<^sub>2)" by blast
      then show ?thesis
      proof (cases)
        case (improp)
        from improp have t': "t = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using t by blast
        from improp have t: "t = l \<cdot> \<sigma>" "t = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
          using \<open>l |_ p \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> subt_at.simps(1) tp' by auto
        from improp obtain \<pi>' where "\<pi>' \<bullet> ?\<rho> = ?\<rho>\<^sub>2" by auto
        then have ren: "(\<pi>' + \<pi>) \<bullet> ?\<rho>\<^sub>1 = ?\<rho>\<^sub>2"
          (is "?\<pi> \<bullet> ((_, _), _) = _") using rename by (auto simp: eqvt)
        then have ren': "l\<^sub>2 = ?\<pi> \<bullet> l\<^sub>1" "r\<^sub>2 = ?\<pi> \<bullet> r\<^sub>1" "cs\<^sub>2 = ?\<pi> \<bullet> cs\<^sub>1" by (auto simp: eqvt)
        have ren'': "l\<^sub>1 = -?\<pi> \<bullet> l\<^sub>2" "r\<^sub>1 = -?\<pi> \<bullet> r\<^sub>2" "r\<^sub>1 = -?\<pi> \<bullet> r\<^sub>2"
          by (auto simp: ren') (metis term_pt.permute_minus_cancel(2) term_pt.permute_plus)+
        have ren_cs: "cs\<^sub>1 = -?\<pi> \<bullet> cs\<^sub>2" using ren' by (metis rules_pt.permute_minus_cancel(2))
        have l2subst: "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = l\<^sub>2 \<cdot> (sop (-?\<pi>) \<circ>\<^sub>s \<sigma>\<^sub>1)" (is "_ = _ \<cdot> ?\<sigma>\<^sub>1")
          using t t' ren'' by (auto simp: eqvt)
        from this [symmetric] have substl: "\<forall>x \<in> vars_term l\<^sub>2. ?\<sigma>\<^sub>1 x = \<sigma>\<^sub>2 x"
          using term_subst_eq_rev by blast
  
        have "t = l\<^sub>2 \<cdot> ?\<sigma>\<^sub>1" using \<open>l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = l\<^sub>2 \<cdot> sop (- (\<pi>' + \<pi>)) \<circ>\<^sub>s \<sigma>\<^sub>1\<close> t(2) by blast
        have t1': "t\<^sub>1 = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" by (metis C C' EQ ctxt_apply_term.simps(1) ctxt_supt supt_neqD tq')
        from t1' have t1: "t\<^sub>1 = r\<^sub>2 \<cdot> ?\<sigma>\<^sub>1" using ren'' by simp
        from improp have t2: "t\<^sub>2 = r\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (metis D' p subt_at.simps(1) subt_at_hole_pos)

        obtain \<sigma>\<^sub>1' where norm_\<sigma>\<^sub>1': "normalized R \<sigma>\<^sub>1'" and \<sigma>\<^sub>1': "\<forall>x. (?\<sigma>\<^sub>1 x, \<sigma>\<^sub>1' x) \<in> (cstep R)\<^sup>*"
          and \<sigma>\<^sub>1'_def: "\<sigma>\<^sub>1' = (\<lambda>x. some_NF (cstep R) (?\<sigma>\<^sub>1 x))"
          using quasi_decreasing_obtains_normalized_subst [OF S] by auto
        obtain \<sigma>\<^sub>2' where norm_\<sigma>\<^sub>2': "normalized R \<sigma>\<^sub>2'" and \<sigma>\<^sub>2': "\<forall>x. (\<sigma>\<^sub>2 x, \<sigma>\<^sub>2' x) \<in> (cstep R)\<^sup>*"
          and \<sigma>\<^sub>2'_def: "\<sigma>\<^sub>2' = (\<lambda>x. some_NF (cstep R) (\<sigma>\<^sub>2 x))"
          using quasi_decreasing_obtains_normalized_subst [OF S] by auto

        have "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
          using cs2 by (auto simp: eqvt)
        then have cs2': "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
          using csteps_n_subset_csteps by force
        have "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> ?\<sigma>\<^sub>1, t \<cdot> ?\<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
          using cs1 [unfolded ren_cs] by (auto simp: eqvt)
        then have cs2'': "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> ?\<sigma>\<^sub>1, t \<cdot> ?\<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
          using csteps_n_subset_csteps by blast
  
        { fix x
          from quasi_decreasing_order_conds_sat [OF S r2 cs2'']
            have 1: "\<forall>i < length cs\<^sub>2. (fst (cs\<^sub>2 ! i) \<cdot> ?\<sigma>\<^sub>1, l\<^sub>2 \<cdot> ?\<sigma>\<^sub>1) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" by auto
          from quasi_decreasing_order_conds_sat [OF S r2 cs2']
            have 2: "\<forall>i < length cs\<^sub>2. (fst (cs\<^sub>2 ! i) \<cdot> \<sigma>\<^sub>2, l\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" by auto
          from SDTRS [unfolded sdtrs_def strongly_deterministic_def] r2
            have 3: "\<forall>(u, v) \<in> set cs\<^sub>2. strongly_irreducible R v" by fast
  
          let ?det = "\<lambda>x. \<forall>i < length x. vars_term (fst (x ! i)) \<subseteq> vars_term l\<^sub>2 \<union> \<Union>(vars_term ` rhss (set (take i x)))"
          assume "x \<in> evars_crule ?\<rho>\<^sub>2"
          moreover from SDTRS [unfolded sdtrs_def wf_ctrs_def dctrs_def X_vars_def] r2
            have "?det cs\<^sub>2" by fastforce
          ultimately have "\<sigma>\<^sub>1' x = \<sigma>\<^sub>2' x" using cs2' cs2'' 1 2 3
          proof (induct cs\<^sub>2 arbitrary: x rule: List.rev_induct)
            case Nil
            then show ?case by (auto simp: evars_crule_def vars_trs_def)
          next
            case (snoc d ds)
            from snoc(3) have det: "?det ds"
              unfolding take_append nth_append by (auto split: if_splits)
            from snoc(4) have sat1: "\<forall>(s,t) \<in> set ds. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
              by fastforce
            from snoc(5) have sat2: "\<forall>(s,t) \<in> set ds. (s \<cdot> ?\<sigma>\<^sub>1, t \<cdot> ?\<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
              by fastforce
            from snoc(6) have qr1: "\<forall>i<length ds. (fst (ds ! i) \<cdot> ?\<sigma>\<^sub>1, l\<^sub>2 \<cdot> ?\<sigma>\<^sub>1) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>"
              unfolding nth_append by (simp split: if_splits add: o_def)
            from snoc(7) have qr2: "\<forall>i<length ds. (fst (ds ! i) \<cdot> \<sigma>\<^sub>2, l\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>"
              unfolding nth_append by (simp split: if_splits add: o_def)
            from snoc(8) have si: "\<forall>(u, v) \<in> set ds. strongly_irreducible R v" by auto
              
            from snoc(1) [OF _ det sat1 sat2 qr1 qr2 si]
              have IH': "\<And>y. y\<in> evars_crule ((l\<^sub>2, r\<^sub>2), ds) \<Longrightarrow> \<sigma>\<^sub>1' y = \<sigma>\<^sub>2' y" by blast
            consider (DS) "x \<in> evars_crule ((l\<^sub>2, r\<^sub>2), ds)"
              | (U) u where "x \<in> vars_term u" "u = fst d"
              | (V) v where "x \<in> vars_term v" "v = snd d"
              using snoc(2) by (auto simp: evars_crule_def vars_trs_def vars_rule_def)
            then show ?case
            proof (cases)
              case (DS)
              from IH' [OF this] show ?thesis .
            next
              case (U)
              then have "x \<in> vars_term (fst ((ds @ [d]) ! (length ds)))" by simp
              with snoc(3) have "x \<in> vars_term l\<^sub>2 \<union> \<Union>(vars_term ` rhss (set (take (length ds) (ds @ [d]))))"
                by fastforce
              with snoc(2) have "x \<in> evars_crule ((l\<^sub>2, r\<^sub>2), ds)"
                by (auto simp: evars_crule_def vars_defs)
              from IH' [OF this] show ?thesis .
            next
              case (V)
              let ?u = "fst d"
              { fix y
                assume "y \<in> vars_term ?u"
                then have "y \<in> vars_term (fst ((ds @ [d]) ! (length ds)))" by simp
                with snoc(3) have "y \<in> vars_term l\<^sub>2 \<union>
                  \<Union>(vars_term ` rhss (set (take (length ds) (ds @ [d]))))"
                  by fastforce
                then consider (L) "y \<in> vars_term l\<^sub>2" | (E) "y \<in> evars_crule ((l\<^sub>2, r\<^sub>2), ds)"
                  by (auto simp: evars_crule_def vars_defs)
                then have "(?\<sigma>\<^sub>1 y, \<sigma>\<^sub>2 y) \<in> (cstep R)\<^sup>\<down>"
                proof (cases)
                  case (L)
                  then have "?\<sigma>\<^sub>1 y = \<sigma>\<^sub>2 y" by (simp add: substl)
                  then show ?thesis by auto
                next
                  case (E)
                  from IH' [OF E] have "\<sigma>\<^sub>1' y = \<sigma>\<^sub>2' y" .
                  then show ?thesis using \<sigma>\<^sub>1' \<sigma>\<^sub>2' joinI by auto
                qed
              }
              then have "\<And>y. y \<in> vars_term ?u \<Longrightarrow> (?\<sigma>\<^sub>1 y, \<sigma>\<^sub>2 y) \<in> (cstep R)\<^sup>\<down>" by auto
              with term_subst_csteps_join have "(?u \<cdot> ?\<sigma>\<^sub>1, ?u \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>\<down>" by blast
              then obtain s\<^sub>0 where u1: "(?u \<cdot> ?\<sigma>\<^sub>1, s\<^sub>0) \<in> (cstep R)\<^sup>*"
                and u2: "(?u \<cdot> \<sigma>\<^sub>2, s\<^sub>0) \<in> (cstep R)\<^sup>*" by blast
              have d': "(fst d, snd d) \<in> set (ds @ [d])" by simp
              then have "(?u \<cdot> ?\<sigma>\<^sub>1, v \<cdot> ?\<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*" unfolding V(2) using snoc(5) by blast
              moreover from substs_csteps [OF \<sigma>\<^sub>1' [rule_format]] have "(v \<cdot> ?\<sigma>\<^sub>1, v \<cdot> \<sigma>\<^sub>1') \<in> (cstep R)\<^sup>*" by fast
              finally have u1': "(?u \<cdot> ?\<sigma>\<^sub>1, v \<cdot> \<sigma>\<^sub>1') \<in> (cstep R)\<^sup>*" by blast
              have "(?u \<cdot> ?\<sigma>\<^sub>1, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" using snoc(6) unfolding \<open>t = l\<^sub>2 \<cdot> ?\<sigma>\<^sub>1\<close>
                by (metis in_set_conv_nth prod.collapse d')
              from IH [OF this u1' u1] have "(v \<cdot> \<sigma>\<^sub>1', s\<^sub>0) \<in> (cstep R)\<^sup>\<down>" .
              then obtain s\<^sub>1 where s1': "(v \<cdot> \<sigma>\<^sub>1', s\<^sub>1) \<in> (cstep R)\<^sup>*" and s1: "(s\<^sub>0, s\<^sub>1) \<in> (cstep R)\<^sup>*" by auto
              have *: "(?u \<cdot> \<sigma>\<^sub>2, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" using snoc(7) unfolding \<open>t = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close>
                by (metis in_set_conv_nth prod.collapse d')
              from u2 s1 have u2': "(fst d \<cdot> \<sigma>\<^sub>2, s\<^sub>1) \<in> (cstep R)\<^sup>*" by auto
              have "(?u \<cdot> \<sigma>\<^sub>2, v \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*" unfolding V(2) using snoc(4) d' by blast
              moreover from substs_csteps [OF \<sigma>\<^sub>2' [rule_format]] have "(v \<cdot> \<sigma>\<^sub>2, v \<cdot> \<sigma>\<^sub>2') \<in> (cstep R)\<^sup>*" by fast
              finally have u2'': "(?u \<cdot> \<sigma>\<^sub>2, v \<cdot> \<sigma>\<^sub>2') \<in> (cstep R)\<^sup>*" by blast
              from IH [OF * u2' u2''] have 2: "(s\<^sub>1, v \<cdot> \<sigma>\<^sub>2') \<in> (cstep R)\<^sup>\<down>" .
              have 5: "(v \<cdot> \<sigma>\<^sub>1', v \<cdot> \<sigma>\<^sub>2') \<in> (cstep R)\<^sup>\<down>" using 2 s1' rtrancl_join_join by metis
              from snoc(8) d' V(2) have siv: "strongly_irreducible R v" by auto
              from norm_\<sigma>\<^sub>1' siv have 3: "v \<cdot> \<sigma>\<^sub>1' \<in> NF (cstep R)" by (auto simp: strongly_irreducible_def)
              from norm_\<sigma>\<^sub>2' siv have 4: "v \<cdot> \<sigma>\<^sub>2' \<in> NF (cstep R)" by (auto simp: strongly_irreducible_def)
              from 3 4 5 show ?thesis by (meson V(1) join_NF_imp_eq term_subst_eq_rev)
            qed
          qed
        }
        then have evars_join: "\<forall>x \<in> evars_crule ?\<rho>\<^sub>2. \<sigma>\<^sub>1' x = \<sigma>\<^sub>2' x" by auto
        from SDTRS [unfolded sdtrs_def wf_ctrs_def type3_def] r2
          have "vars_term r\<^sub>2 \<subseteq> vars_term l\<^sub>2 \<union> vars_trs (set cs\<^sub>2)"
          using fst_conv snd_conv by fastforce
        then have "vars_crule ?\<rho>\<^sub>2 = evars_crule ?\<rho>\<^sub>2 \<union> vars_term l\<^sub>2"
          unfolding evars_crule_def
          proof -
            have "vars_crule ?\<rho>\<^sub>2 = vars_trs (set cs\<^sub>2) \<union> (vars_term l\<^sub>2 \<union> vars_term r\<^sub>2)"
              by (simp add: inf_sup_aci(5) vars_crule_def vars_rule_def)
            then show "vars_crule ?\<rho>\<^sub>2 = vars_trs (set (conds ?\<rho>\<^sub>2)) - vars_term (clhs ?\<rho>\<^sub>2) \<union> vars_term l\<^sub>2"
              using \<open>vars_term r\<^sub>2 \<subseteq> vars_term l\<^sub>2 \<union> vars_trs (set cs\<^sub>2)\<close> snd_conv by auto
          qed
        then have "\<forall>x \<in> vars_crule ?\<rho>\<^sub>2. (\<sigma>\<^sub>1' x, \<sigma>\<^sub>2' x) \<in> (cstep R)\<^sup>\<down>"
          using evars_join substl \<sigma>\<^sub>1' \<sigma>\<^sub>2' \<sigma>\<^sub>1'_def \<sigma>\<^sub>2'_def by auto
        then have "\<forall>x \<in> vars_crule ?\<rho>\<^sub>2. (?\<sigma>\<^sub>1 x, \<sigma>\<^sub>2 x) \<in> (cstep R)\<^sup>\<down>"
          by (meson \<sigma>\<^sub>1' \<sigma>\<^sub>2' join_rtrancl_join rtrancl_join_join)
        then show ?thesis unfolding t1 t2
          by (metis Un_iff fst_conv prod.sel(2) term_subst_csteps_join vars_crule_def vars_defs(2))
      next
        case (proper)
        def \<sigma>' \<equiv> "\<lambda>x. if x \<in> vars_crule ?\<rho> then \<sigma> x else \<sigma>\<^sub>2 x"
        have unif: "l\<^sub>2 \<cdot> \<sigma>' = (l |_ p) \<cdot> \<sigma>'"
        proof -
          note coinc = coincidence_lemma' [of l\<^sub>2 "vars_crule ?\<rho>\<^sub>2"]
          have disj: "vars_crule ?\<rho>\<^sub>2 \<inter> vars_term (l |_ p) = {}"
            using vars_term_subt_at [OF poss] and disj
            by (auto simp: vars_crule_def vars_rule_def)
          have "l\<^sub>2 \<cdot> \<sigma>' = l\<^sub>2 \<cdot> (\<sigma>' |s vars_crule ?\<rho>\<^sub>2)"
            using coinc by (simp add: vars_crule_def vars_rule_def sup_assoc)
          also have "\<dots> = l\<^sub>2 \<cdot> (\<sigma>\<^sub>2 |s vars_crule ?\<rho>\<^sub>2)" apply (simp add: \<sigma>'_def)
            by (metis (no_types) \<pi> crule_pt.permute_prod.simps cs_def l_def r_def
                rule_pt.permute_prod.simps subst_restrict_restrict)
          also have "\<dots> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using coinc by (simp add: vars_crule_def vars_rule_def) blast
          also have "\<dots> = (l \<cdot> \<sigma>) |_ p" by (simp add: \<open>l |_ p \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> poss)
          also have "\<dots> = (l |_ p) \<cdot> \<sigma>" by (simp add: \<open>l |_ p \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> \<open>l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = l \<cdot> \<sigma> |_ p\<close>) 
          also have "\<dots> = (l |_ p) \<cdot> (\<sigma> |s vars_term (l |_ p))"
            by (simp add: coincidence_lemma [symmetric])
          also have "\<dots> = (l |_ p) \<cdot> (\<sigma>' |s vars_term (l |_ p))" using disj
            apply (simp add: \<sigma>'_def)
            by (metis (no_types) \<open>vars_term (l |_ p) \<subseteq> vars_term l\<close> coincidence_lemma coincidence_lemma' fst_conv subst_restrict_vars sup.coboundedI1 vars_crule_def vars_rule_def)
          finally show ?thesis by (simp add: coincidence_lemma [symmetric])
        qed
        from is_imgu_imp_is_mgu [OF mgu_sound [OF mgu]] have "is_mgu \<tau> {(l |_ p, l\<^sub>2)}" .
        with unif obtain \<delta> where \<sigma>': "\<sigma>' = \<tau> \<circ>\<^sub>s \<delta>"
          by (auto simp: is_mgu_def unifiers_def) presburger
        from proper CCP_I [OF rule_variants disj funposs] and mgu
          have ccp: "((r \<cdot> \<tau>, (ctxt_of_pos_term p l)\<langle>r\<^sub>2\<rangle> \<cdot> \<tau>), subst_list \<tau> (cs @ cs\<^sub>2)) \<in> CCP R"
          by simp
        have sig:  "\<forall>(s, t) \<in> set cs. \<forall>x \<in> vars_term s \<union> vars_term t. \<sigma> x = \<sigma>' x"
          by (fastforce simp: \<sigma>'_def vars_crule_def vars_trs_def vars_rule_def)
        { fix s t
          assume a: "(s, t) \<in> set cs"
          with sig have "\<forall>x \<in> vars_term s. \<sigma> x = \<sigma>' x" and "\<forall>x \<in> vars_term t. \<sigma> x = \<sigma>' x"
            by auto
          moreover with cs a have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" by fast
          ultimately have "(s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*"
            using term_subst_eq by metis
        }
        then have 1: "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*" by auto
        have sig2:  "\<forall>(s, t) \<in> set cs\<^sub>2. \<forall>x \<in> vars_term s \<union> vars_term t. \<sigma>\<^sub>2 x = \<sigma>' x"
          using disj by (fastforce simp: \<sigma>'_def vars_crule_def vars_trs_def vars_rule_def)
        { fix s t
          assume a: "(s, t) \<in> set cs\<^sub>2"
          with sig2 have "\<forall>x \<in> vars_term s. \<sigma>\<^sub>2 x = \<sigma>' x" and "\<forall>x \<in> vars_term t. \<sigma>\<^sub>2 x = \<sigma>' x"
            by auto
          moreover with cs2 a have "(s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*" by fast
          ultimately have "(s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*"
            using term_subst_eq by metis
        }
        then have 2: "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*" by auto

        have t1: "t\<^sub>1 = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" by (metis C C' ctxt_apply_term.simps(1) ctxt_supt
            subterm.dual_order.strict_implies_not_eq t)
        from t tp' have "t = (ctxt_of_pos_term p (l\<^sub>1 \<cdot> \<sigma>\<^sub>1))\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
          by (metis \<alpha> ctxt_supt_id funposs_imp_poss poss_imp_subst_poss)
        then have **: "D = ctxt_of_pos_term p (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using p D
          using ctxt_of_pos_term_hole_pos t by force
        then have t2: "t\<^sub>2 = (ctxt_of_pos_term p (l\<^sub>1 \<cdot> \<sigma>\<^sub>1))\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
          using D' [unfolded **] by simp
        from rename have "\<sigma>\<^sub>1 = sop \<pi> \<circ>\<^sub>s \<sigma>" unfolding subst_compose_def o_def
          by (metis inv_Rep_perm_simp permute_atom_def subst_apply_term.simps(1))
        from t1 rename have "t\<^sub>1 = (-\<pi>) \<bullet> r \<cdot> (sop \<pi> \<circ>\<^sub>s \<sigma>)" by force
        then have "t\<^sub>1 = r \<cdot> \<sigma>" by simp
        then have t1': "t\<^sub>1 = r \<cdot> \<sigma>'" using \<sigma>'_def
          by (simp add: UnCI term_subst_eq_conv vars_crule_def vars_defs(2))
        from t2 rename have "t\<^sub>2 = (ctxt_of_pos_term p (l \<cdot> \<sigma>))\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>" by force
        moreover have "l \<cdot> \<sigma> = l \<cdot> \<sigma>'" using \<sigma>'_def
          by (simp add: UnCI term_subst_eq_conv vars_crule_def vars_defs(2))
        moreover have "r\<^sub>2 \<cdot> \<sigma>\<^sub>2 = r\<^sub>2 \<cdot> \<sigma>'" using disj
          by (induct r\<^sub>2) (auto simp: \<sigma>'_def vars_crule_def vars_rule_def)
        ultimately have "t\<^sub>2 = (ctxt_of_pos_term p (l \<cdot> \<sigma>'))\<langle>r\<^sub>2 \<cdot> \<sigma>'\<rangle>" by presburger
        then have t2': "t\<^sub>2 = (ctxt_of_pos_term p l)\<langle>r\<^sub>2\<rangle> \<cdot> \<sigma>'"
          by (simp add: ctxt_of_pos_term_subst poss)

        from 1 2 have "\<forall>(s, t) \<in> set (cs @ cs\<^sub>2). (s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in> (cstep R)\<^sup>*"
          using UnE set_append by auto
        then obtain n where "conds_sat R n (cs @ cs\<^sub>2) \<sigma>'" using conds_sat_iff' by blast
        from conds_sat_subst_list [THEN iffD2, OF this [unfolded \<sigma>']]
          have "conds_sat R n (subst_list \<tau> (cs @ cs\<^sub>2)) \<delta>" .
        (* difference to 4.1 start *)
        (* changes from Theorem 4.1 to 4.2 are here:
           "The only point we have to consider is the first case (proper) in (\<alpha>):" *)
        then have "\<exists>n. conds_sat R n (subst_list \<tau> (cs @ cs\<^sub>2)) \<delta>" by auto
        from conds_sat_iff' [THEN iffD1, OF this]
          have sat: "\<forall>(s, t) \<in> set (subst_list \<tau> (cs @ cs\<^sub>2)). (s \<cdot> \<delta>, t \<cdot> \<delta>) \<in> (cstep R)\<^sup>*" .
        let ?C = "set (subst_list \<tau> (cs @ cs\<^sub>2))"
        consider (unf) "unfeasible R S ?\<rho> ?\<rho>\<^sub>2 p"
               | (cjn) "context_joinable R ?\<rho> ?\<rho>\<^sub>2 p"
               using allCPs ccp by (metis * \<open>0 \<bullet> ?\<rho>\<^sub>2 \<in> R\<close> disj fst_conv funposs mgu overlapI)
        then show ?thesis
        proof (cases)
          case (unf)
          then obtain t\<^sub>0 t\<^sub>0' t\<^sub>0''
            where 1: "(clhs ?\<rho> \<cdot> \<tau>, t\<^sub>0) \<in> S"
            and 2: "(t\<^sub>0, t\<^sub>0') \<in> (ctxt_step (cstep R) ?C)\<^sup>*"
            and 3: "(t\<^sub>0, t\<^sub>0'') \<in> (ctxt_step (cstep R) ?C)\<^sup>*"
            and 6: "\<not> unifiable {(t\<^sub>0', t\<^sub>0'')}"
            and 7: "strongly_irreducible R t\<^sub>0'"
            and 8: "strongly_irreducible R t\<^sub>0''"
            unfolding unfeasible_def by (metis fst_conv mgu option.sel snd_conv)
          from 1 have 9: "(clhs ?\<rho> \<cdot> \<tau> \<cdot> \<delta>, t\<^sub>0 \<cdot> \<delta>) \<in> (S \<union> {\<rhd>})\<^sup>+" using subst
            by (auto simp: subst_closed_supt)
          from ctxt_steps_csteps [OF 2 sat] have 4: "(t\<^sub>0 \<cdot> \<delta>, t\<^sub>0' \<cdot> \<delta>) \<in> (cstep R)\<^sup>*" .
          from ctxt_steps_csteps [OF 3 sat] have 5: "(t\<^sub>0 \<cdot> \<delta>, t\<^sub>0'' \<cdot> \<delta>) \<in> (cstep R)\<^sup>*" .
          
          obtain \<mu> where norm_\<mu>: "normalized R \<mu>" and \<mu>: "\<forall>x. (\<delta> x, \<mu> x) \<in> (cstep R)\<^sup>*"
            using quasi_decreasing_obtains_normalized_subst [OF S] by auto

          have 10: "(t\<^sub>0 \<cdot> \<delta>, t\<^sub>0' \<cdot> \<mu>) \<in> (cstep R)\<^sup>*" by (meson \<mu> 4 rtrancl_trans substs_csteps)
          have 11: "(t\<^sub>0 \<cdot> \<delta>, t\<^sub>0'' \<cdot> \<mu>) \<in> (cstep R)\<^sup>*" by (meson \<mu> 5 rtrancl_trans substs_csteps)
          have 12: "t\<^sub>0' \<cdot> \<mu> \<noteq> t\<^sub>0'' \<cdot> \<mu>" using 6 unfolding unifiable_def unifiers_def by auto
          
          have "t = l \<cdot> \<sigma>" using t by (simp add: rename)
          also have "\<dots> = l \<cdot> \<sigma>'" using \<sigma>'_def by (simp add: term_subst_eq_conv vars_crule_def vars_rule_def)
          finally have "t = l \<cdot> \<tau> \<cdot> \<delta>" using \<sigma>' by simp
          with 9 have "(t\<^sub>0 \<cdot> \<delta>, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" by auto
          from IH [OF this 10 11] have "(t\<^sub>0' \<cdot> \<mu>, t\<^sub>0'' \<cdot> \<mu>) \<in> (cstep R)\<^sup>\<down>" .
          with 7 8 12 show ?thesis by (simp add: strongly_irreducible_def norm_\<mu> join_NF_imp_eq)
        next
          let ?s\<^sub>1 =  "crhs ?\<rho> \<cdot> \<tau>"
          let ?s\<^sub>2 = "(ctxt_of_pos_term p (clhs ?\<rho>))\<langle>crhs ?\<rho>\<^sub>2\<rangle> \<cdot> \<tau>"
          case (cjn)
          then obtain t\<^sub>0
            where 1: "(?s\<^sub>1, t\<^sub>0) \<in> (ctxt_step (cstep R) ?C)\<^sup>*"
            and 2: "(?s\<^sub>2, t\<^sub>0) \<in> (ctxt_step (cstep R) ?C)\<^sup>*"
            unfolding context_joinable_def by (metis fst_conv mgu option.sel snd_conv)
          from ctxt_steps_csteps [OF 1 sat] ctxt_steps_csteps [OF 2 sat]
            have 3: "(?s\<^sub>1 \<cdot> \<delta>, ?s\<^sub>2 \<cdot> \<delta>) \<in> (cstep R)\<^sup>\<down>" (is "(?t\<^sub>1, ?t\<^sub>2) \<in> _") by blast

          have "t\<^sub>1 = ?t\<^sub>1" using \<sigma>' t1' by simp
          moreover have "t\<^sub>2 = ?t\<^sub>2" using \<sigma>' t2' by simp
          ultimately show ?thesis using 3 by simp
        qed
        (* difference to 4.1 end *)
      qed
    next
      case (\<beta>)
      moreover have \<beta>':  "p \<in> poss (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using D p t by auto
      from poss_subst_apply_term [OF \<beta>' \<beta>]
        obtain q\<^sub>1 q\<^sub>2 x where [simp]: "p = q\<^sub>1 <#> q\<^sub>2" and q\<^sub>1: "q\<^sub>1 \<in> poss l\<^sub>1"
      and l\<^sub>1q\<^sub>1: "l\<^sub>1 |_ q\<^sub>1 = Var x" and q\<^sub>2: "q\<^sub>2 \<in> poss (\<sigma>\<^sub>1 x)" by auto
      moreover
        have [simp]: "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 |_ p = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using tp' [unfolded t] .
      ultimately
        have [simp]: "\<sigma>\<^sub>1 x |_ q\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" by simp
  
      have \<sigma>1step: "(\<sigma>\<^sub>1 x, replace_at (\<sigma>\<^sub>1 x) q\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) \<in> cstep R" (is "(_, ?t\<^sub>0) \<in> _")
        using cstep_ctxt s2 by (metis \<open>\<sigma>\<^sub>1 x |_ q\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> ctxt_supt_id q\<^sub>2)
      
      def \<tau> \<equiv> "\<lambda>y. if y = x then ?t\<^sub>0 else \<sigma>\<^sub>1 y"

      obtain \<tau>' where norm_\<tau>': "normalized R \<tau>'" and \<tau>': "\<forall>x. (\<tau> x, \<tau>' x) \<in> (cstep R)\<^sup>*"
        using quasi_decreasing_obtains_normalized_subst [OF S] by auto
      
      have "(\<sigma>\<^sub>1 x, \<tau> x) \<in> cstep R"
      proof -
        let ?C = "ctxt_of_pos_term q\<^sub>2 (\<sigma>\<^sub>1 x)"
        have "(?C\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>, ?C\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>) \<in> cstep R" using s2 by (simp add: cstep_ctxt) 
        then show ?thesis using q\<^sub>2 by (simp add: \<tau>_def replace_at_ident)
      qed
      then have *: "\<And>x. (\<sigma>\<^sub>1 x, \<tau> x) \<in> (cstep R)\<^sup>*" by (auto simp: \<tau>_def)
  
      have "(t\<^sub>2, l\<^sub>1 \<cdot> \<tau>) \<in> (cstep R)\<^sup>*"
      proof -
        have "replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2) = replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) q\<^sub>1 (\<tau> x)"
          using q\<^sub>1 and q\<^sub>2 by (simp add: \<tau>_def ctxt_of_pos_term_append l\<^sub>1q\<^sub>1)
        moreover
        have "(replace_at (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) q\<^sub>1 (\<tau> x), l\<^sub>1 \<cdot> \<tau>) \<in> (cstep R)\<^sup>*"
          by (rule replace_at_subst_csteps [OF * q\<^sub>1 l\<^sub>1q\<^sub>1])
        then show ?thesis
          by (metis D D' calculation ctxt_of_pos_term_hole_pos p t)
      qed
      then have t2csteps: "(t\<^sub>2, l\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" unfolding t using \<tau>'
        by (meson * rtrancl_trans substs_csteps)
      
      have "\<forall>x \<in> vars_term r\<^sub>1. (\<sigma>\<^sub>1 x, \<tau> x) \<in> (cstep R)\<^sup>*" using * \<sigma>1step rtrancl.simps by auto
      then have "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<tau>) \<in> (cstep R)\<^sup>*" using term_subst_csteps [of r\<^sub>1 \<sigma>\<^sub>1 \<tau>] by blast
      moreover have "(r\<^sub>1 \<cdot> \<tau>, r\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" using \<tau>' by (simp add: substs_csteps)
      ultimately have "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by auto
      then have t1csteps: "(t\<^sub>1, r\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" using t s1
        by (metis C C' ctxt_apply_term.simps(1) ctxt_supt supt_neqD)
      
      { fix u v
        assume a: "(u, v) \<in> set cs\<^sub>1"
        have cs1': "\<forall>(s, t) \<in> set cs\<^sub>1. (s \<cdot> \<sigma>\<^sub>1, t \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
          using cs1 by (auto simp: csteps_n_imp_csteps)
        from a have "(u \<cdot> \<sigma>\<^sub>1, v \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*" using cs1 by (auto simp: csteps_n_imp_csteps)
        then have 1: "(u \<cdot> \<sigma>\<^sub>1, v \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by (meson * \<tau>' rtrancl_trans substs_csteps)
        
        from quasi_decreasing_order_conds_sat [OF S r1 cs1']
          have 2: "\<forall>i < length cs\<^sub>1. (fst (cs\<^sub>1 ! i) \<cdot> \<sigma>\<^sub>1, l\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" by auto
        have 3: "(u \<cdot> \<sigma>\<^sub>1, u \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by (meson * \<tau>' rtrancl_trans substs_csteps)
        have "(u \<cdot> \<sigma>\<^sub>1, t) \<in> ((S \<union> {\<rhd>})\<^sup>+)\<inverse>" unfolding t using a 2 by (metis fst_conv in_set_idx)
        from IH [OF this 3 1] have 4: "(u \<cdot> \<tau>', v \<cdot> \<tau>') \<in> (cstep R)\<^sup>\<down>" .
        
        from SDTRS [unfolded sdtrs_def strongly_deterministic_def] r1
            have "\<forall>(u, v) \<in> set cs\<^sub>1. strongly_irreducible R v" by fast
        then have siv: "strongly_irreducible R v" using a  by auto
        from norm_\<tau>' siv have "v \<cdot> \<tau>' \<in> NF (cstep R)" by (auto simp: strongly_irreducible_def)
        with 4 have "(u \<cdot> \<tau>', v \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by (simp add: NF_join_imp_reach)
      }
      then have "\<forall>(u, v) \<in> set cs\<^sub>1. (u \<cdot> \<tau>', v \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by auto
      from cstepI [OF r1 this, of "l\<^sub>1 \<cdot> \<tau>'" \<box> "r\<^sub>1 \<cdot> \<tau>'"]
        have "(l\<^sub>1 \<cdot> \<tau>', r\<^sub>1 \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by simp
      with t1csteps t2csteps
       show ?thesis using joinI rtrancl_trans by auto
    qed
  qed
qed

text \<open>Theorem 4.2 from AL94, but using quasi-decreasingness\<close>
theorem quasi_decreasing_order_sdtrs_all_CPs_unfeasible_or_context_joinable_CR:
  assumes sdtrs: "sdtrs R"
    and qdo: "quasi_decreasing_order R S"
    and subst: "subst.closed S"
    and allCPs: "\<forall> r r' p. overlap R r r' p \<longrightarrow> unfeasible R S r r' p \<or> context_joinable R r r' p"
  shows "CR (cstep R)"
proof -
  { fix t t' t''
    from qdo have SNS: "SN S" by (auto simp: quasi_decreasing_order_def)
    then have "SN ((S \<union> {\<rhd>})\<^sup>+)" (is "SN ?S")
    proof -
      have "{\<rhd>} \<subseteq> S"
        using qdo quasi_decreasing_order_def by blast
      then show ?thesis
        by (simp add: SNS SN_imp_SN_trancl subset_Un_eq sup_commute)
    qed
    from qdo have trans: "trans S" and RsubS: "cstep R \<subseteq> S"
      by (auto simp: quasi_decreasing_order_def)
    then have *: "cstep R \<subseteq> ?S" using qdo quasi_decreasing_order_def by auto

    assume 1: "(t, t') \<in> (cstep R)\<^sup>*" and 2: "(t, t'') \<in> (cstep R)\<^sup>*"
    then have "(t', t'') \<in> (cstep R)\<^sup>\<down>"
    proof (induct t arbitrary: t' t'' rule: wf_induct[OF SN_imp_wf[OF \<open>SN ?S\<close>], rule_format])
      case (1 t)
      note IH = this(1)
      { assume "t = t' \<or> t = t''"
        then have "(t', t'') \<in> (cstep R)\<^sup>\<down>" using 1 2 by blast
      }
      moreover
      { assume "\<exists>t\<^sub>1. (t, t\<^sub>1) \<in> cstep R \<and> (t\<^sub>1, t') \<in> (cstep R)\<^sup>*"
          and "\<exists>t\<^sub>2. (t, t\<^sub>2) \<in> cstep R \<and> (t\<^sub>2, t'') \<in> (cstep R)\<^sup>*"
        then obtain t\<^sub>1 t\<^sub>2 where t11: "(t, t\<^sub>1) \<in> cstep R" and t12:"(t\<^sub>1, t') \<in> (cstep R)\<^sup>*"
          and t21: "(t, t\<^sub>2) \<in> cstep R" and t22: "(t\<^sub>2, t'') \<in> (cstep R)\<^sup>*" by blast
        from t11 obtain n' where "(t, t\<^sub>1) \<in> cstep_n R n'" using cstep_iff by blast 
        from cstep_nE [OF this] obtain C l\<^sub>1 r\<^sub>1 \<sigma>\<^sub>1 cs\<^sub>1 n
          where r1: "((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R"
          and "n' = Suc n"
          and cs1: "\<forall>(s, t) \<in> set cs\<^sub>1. (s \<cdot> \<sigma>\<^sub>1, t \<cdot> \<sigma>\<^sub>1) \<in> (cstep_n R n)\<^sup>*"
          and C: "t = C\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>"
          and C': "t\<^sub>1 = C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>" by metis
        from t21 obtain m' where "(t, t\<^sub>2) \<in> cstep_n R m'" using cstep_iff by blast
        from cstep_nE [OF this] obtain D l\<^sub>2 r\<^sub>2 \<sigma>\<^sub>2 cs\<^sub>2 m
          where r2: "((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R"
          and "m' = Suc m"
          and cs2: "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep_n R m)\<^sup>*"
          and D: "t = D\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
          and D': "t\<^sub>2 = D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>" by metis
        from cstep_n_SucI [OF r1 cs1, of "l\<^sub>1 \<cdot> \<sigma>\<^sub>1" \<box> "r\<^sub>1 \<cdot> \<sigma>\<^sub>1"]
          have s1: "(l\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> cstep R" by (simp add: cstep_n_imp_cstep)
        from cstep_n_SucI [OF r2 cs2, of "l\<^sub>2 \<cdot> \<sigma>\<^sub>2" \<box> "r\<^sub>2 \<cdot> \<sigma>\<^sub>2"]
          have s2: "(l\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> cstep R" by (simp add: cstep_n_imp_cstep)
        obtain q p where q: "hole_pos C = q" and p: "hole_pos D = p" by simp
        have tq': "t |_q = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using C q by fastforce
        from cs1 have cs1': "\<forall>(s, t) \<in> set cs\<^sub>1. (s \<cdot> \<sigma>\<^sub>1, t \<cdot> \<sigma>\<^sub>1) \<in> (cstep R)\<^sup>*"
          using csteps_n_imp_csteps by blast
        from cs2 have cs2': "\<forall>(s, t) \<in> set cs\<^sub>2. (s \<cdot> \<sigma>\<^sub>2, t \<cdot> \<sigma>\<^sub>2) \<in> (cstep R)\<^sup>*"
          using csteps_n_imp_csteps by blast
        consider (GE) "q \<le> p" | (LE) "p \<le> q" | (PL) "q \<bottom> p" using parallel_pos by blast 
        then have "(t\<^sub>1, t\<^sub>2) \<in> (cstep R)\<^sup>\<down>"
        proof (cases)
          case (GE)
          from aux_lemma' [OF qdo subst allCPs sdtrs r1 cs1' C C' r2 cs2' D D' q p GE IH]
            show ?thesis .
        next
          case (LE)
          from aux_lemma' [OF qdo subst allCPs sdtrs r2 cs2' D D' r1 cs1' C C' p q LE IH]
            have "(t\<^sub>2, t\<^sub>1) \<in> (cstep R)\<^sup>\<down>" .
          then show ?thesis by auto
        next
          case (PL)
          from C D q p have tpar: "replace_at t q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) = replace_at t p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)"
            using hole_pos_id_ctxt by fastforce
          have *: "t = replace_at (replace_at t q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" using C tpar q by auto
          have **: "t = replace_at (replace_at t p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using C tpar q by auto
          then have "C\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle> = replace_at (replace_at t q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" using * C by blast
          then have "C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle> = replace_at (replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)"
            by (metis C D PL ctxt_of_pos_term_hole_pos hole_pos_poss p parallel_replace_at q)
          then have t1: "t\<^sub>1 = replace_at (replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" using C' by blast
          have 1: "(t\<^sub>1, replace_at (replace_at t q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)) p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) \<in> cstep R"
            (is "(_, ?t4) \<in> _") unfolding t1 using s2 by (simp add: cstep_ctxt)
          from ** have "D\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle> = replace_at (replace_at t p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using D by blast
          then have "D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle> = replace_at (replace_at t p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)"
            using D p tpar by (metis C PL ctxt_of_pos_term_hole_pos hole_pos_poss parallel_replace_at q)
          then have t2: "t\<^sub>2 = replace_at (replace_at t p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using D' by blast
          have 2: "(t\<^sub>2, replace_at (replace_at t p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)) q (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)) \<in> cstep R"
            (is "(_, ?t3) \<in> _") unfolding t2 using s1 by (simp add: cstep_ctxt)
          have "?t3 = ?t4" using PL by (metis C D hole_pos_poss p parallel_replace_at q)
          then show ?thesis using 1 2 by auto
        qed
        then obtain s where s01: "(t\<^sub>1, s) \<in> (cstep R)\<^sup>*" and s02: "(t\<^sub>2, s) \<in> (cstep R)\<^sup>*" by blast
        from t11 * have "(t\<^sub>1, t) \<in> ?S\<inverse>" by blast
        from IH [OF this s01 t12] have "(s, t') \<in> (cstep R)\<^sup>\<down>" .
        then obtain s' where s11: "(t', s') \<in> (cstep R)\<^sup>*" and s12: "(s, s') \<in> (cstep R)\<^sup>*" by blast
        from t21 * have t2tqr: "(t\<^sub>2, t) \<in> ?S\<inverse>" by blast
        have "(t\<^sub>2, s') \<in> (cstep R)\<^sup>*" using s02 s12 by force
        from IH [OF t2tqr this t22] have "(s', t'') \<in> (cstep R)\<^sup>\<down>" .
        with s11 have "(t', t'') \<in> (cstep R)\<^sup>\<down>" by (simp add: rtrancl_join_join)
      }
      ultimately show "(t', t'') \<in> (cstep R)\<^sup>\<down>" using 1 2 converse_rtranclE by metis
    qed
  }
  then show "CR (cstep R)" unfolding CR_on_def by force
qed

text \<open>Theorem 4.2 from AL94\<close>
theorem quasi_reductive_order_sdtrs_all_CPs_unfeasible_or_context_joinable_CR:
  assumes sdtrs: "sdtrs R"
    and qro: "quasi_reductive_order R S"
    and subst: "subst.closed S"
    and allCPs: "\<forall> r r' p. overlap R r r' p \<longrightarrow> unfeasible R S r r' p \<or> context_joinable R r r' p"
  shows "CR (cstep R)"
proof -
  from quasi_reductive_order_quasi_decreasing_order [OF qro]
    have qdo: "quasi_decreasing_order R ((S \<union> {\<rhd>})\<^sup>+)" .
  moreover have "subst.closed ((S \<union> {\<rhd>})\<^sup>+)"
    by (simp add: subst subst.closed_trancl subst.closed_Un subst_closed_supt)
  moreover have "\<forall> r r' p. overlap R r r' p \<longrightarrow> unfeasible R ((S \<union> {\<rhd>})\<^sup>+) r r' p \<or> context_joinable R r r' p"
    using allCPs unfeasible_supt_trancl by blast
  ultimately show ?thesis using quasi_decreasing_order_sdtrs_all_CPs_unfeasible_or_context_joinable_CR sdtrs
    by blast
qed

end
