theory Ordered_Algebra
  imports Misc "../Orderings/Reduction_Pair" "~~/HOL/Lattices" "~~/HOL/Library/Polynomial"
begin

section{* Quasi-Ordered Domains *}

class quasi_order = ord +
  assumes order_refl[iff]: "x \<le> x"
      and order_trans[trans]: "x \<le> y \<Longrightarrow> y \<le> z \<Longrightarrow> x \<le> z"
      and less_trans[trans]: "x < y \<Longrightarrow> y < z \<Longrightarrow> x < z"
      and le_less_trans[trans]: "x \<le> y \<Longrightarrow> y < z \<Longrightarrow> x < z"
      and less_le_trans[trans]: "x < y \<Longrightarrow> y \<le> z \<Longrightarrow> x < z"
begin

text {*
 Class @{class quasi_order} is incomparable with HOL class @{class preorder}, as the latter
 fixes $<$ as the "strict part" (which is not nice, e.g., for comparing polynomials)
 while we already assume compatibility (@{thm le_less_trans}).
*}

definition equiv (infix "\<sim>" 50) where [iff]: "x \<sim> y \<equiv> x \<le> y \<and> y \<le> x"

definition equiv_class ("[_]\<^sub>\<sim>") where "[x]\<^sub>\<sim> \<equiv> { y. x \<sim> y }"

lemma mem_equiv_class[simp]: "y \<in> [x]\<^sub>\<sim> \<longleftrightarrow> x \<sim> y" by (auto simp: equiv_class_def)

definition "UB X \<equiv> {b. \<forall>x \<in> X. b \<ge> x}"

lemma UB_I[intro]: "(\<And>x. x \<in> X \<Longrightarrow> b \<ge> x) \<Longrightarrow> b \<in> UB X" by (auto simp: UB_def)

lemma UB_D[dest]: "b \<in> UB X \<Longrightarrow> x \<in> X \<Longrightarrow> b \<ge> x" by (auto simp: UB_def)

lemma mem_UB[simp]: "b \<in> UB X \<longleftrightarrow> (\<forall>x \<in> X. b \<ge> x)" by auto

definition "LUB X \<equiv> { b \<in> UB X. \<forall>c \<in> UB X. b \<le> c }"

lemma LUB_I[intro]: assumes "b \<in> UB X" "\<And>c. c \<in> UB X \<Longrightarrow> b \<le> c" shows "b \<in> LUB X"
  using assms by (auto simp: LUB_def)

lemma LUB_D[dest]: assumes "b \<in> LUB X" shows "b \<in> UB X" "c \<in> UB X \<Longrightarrow> b \<le> c"
  using assms[unfolded LUB_def] by auto

lemma LUB_singleton: "LUB {x} = [x]\<^sub>\<sim>"
proof(intro equalityI subsetI; unfold mem_equiv_class)
  fix b assume "b \<in> LUB {x}" then have "b \<in> UB {x}" "\<And>c. c \<in> UB {x} \<Longrightarrow> b \<le> c" by auto
  then have "b \<ge> x" "\<And>c. c \<ge> x \<Longrightarrow> b \<le> c" by auto
  then show "x \<sim> b" by auto
next
  fix b assume xb: "x \<sim> b" then show "b \<in> LUB {x}" by(intro LUB_I, auto dest:order_trans)
qed

lemma LUB_equiv: assumes bLUB: "b \<in> LUB X" shows "LUB X = [b]\<^sub>\<sim>"
proof(intro equalityI subsetI, unfold mem_equiv_class)
  from bLUB have bUB: "b \<in> UB X" and leastb: "\<And>x. x \<in> UB X \<Longrightarrow> b \<le> x" by auto
  { fix c assume "c \<in> LUB X"
    then have cUB: "c \<in> UB X" and leastc: "\<And>b. b \<in> UB X \<Longrightarrow> c \<le> b" by auto
    from leastb[OF cUB] leastc[OF bUB] show "b \<sim> c" by auto
  }
  { fix c assume bc: "b \<sim> c"
    show "c \<in> LUB X"
    proof(intro LUB_I UB_I)
      fix x assume "x \<in> X"
      with bUB have "x \<le> b" by auto
      with bc show "x \<le> c" by (auto dest: order_trans)
    next
      fix x assume "x \<in> UB X"
      from leastb[OF this] bc show "c \<le> x" by (auto dest: order_trans)
    qed
  }
qed

definition S where "S \<equiv> {(s,t). s > t}"
definition NS where "NS \<equiv> {(s,t). s \<ge> t}"

sublocale order_pair S NS
  apply (unfold_locales, unfold NS_def S_def, fold transp_trans)
  using le_less_trans less_le_trans
  by (auto intro: refl_onI transpI[OF less_trans] transpI[OF order_trans])

end

subsection {* Weakly ordered domain: where antisymmetry holds *}
class weak_order = quasi_order + assumes antisym[simp]: "a \<le> b \<Longrightarrow> b \<le> a \<Longrightarrow> a = b"
begin

lemma equiv_is_eq[simp]: "op \<sim> = op =" by (intro ext, auto)

lemma equiv_class_singleton[simp]: "[s]\<^sub>\<sim> = {s}" by auto

end

subclass(in order) weak_order by(unfold_locales, auto)


subsection {* Well-founded quasi-ordered domain *}

class wf_order = quasi_order +
  assumes less_induct[case_names less]: "\<And>P a. (\<And>x. (\<And>y. y < x \<Longrightarrow> P y) \<Longrightarrow> P x) \<Longrightarrow> P a"
begin

lemma wf: "wf {(s,t). s < t}" unfolding wf_def by(auto intro:less_induct)

sublocale SN_order_pair S NS
  by (unfold_locales, auto simp: S_def SN_iff_wf converse_def conversep.simps intro: wf)

end

text {* @{class wellorder} is a @{class wf_order}. *}

subclass(in wellorder) wf_order by (unfold_locales; (fact less_induct | simp))

text {* @{type nat} is a @{class wf_order}. *}
instance nat :: wf_order ..

subsection {* Quasi-ordered domain with @{term sup} operation. *}

class quasi_order_sup = quasi_order + sup +
  assumes sup_ge1 [simp]: "x \<le> sup x y"
      and sup_ge2 [simp]: "y \<le> sup x y"
begin

lemma sup_trans1: assumes "x \<le> y" shows "x \<le> sup y z"
proof-
  note assms also note sup_ge1 finally show ?thesis by auto
qed

lemma sup_trans2: assumes "x \<le> z" shows "x \<le> sup y z"
proof-
  note assms also note sup_ge2 finally show ?thesis by auto
qed

lemma finite_set_has_UB: "finite X \<Longrightarrow> UB X \<noteq> {}"
proof(unfold UB_def, induct X rule:finite_induct)
  case empty show ?case by auto
next
  case (insert x X)
  then obtain b where *: "\<And>x. x \<in> X \<Longrightarrow> x \<le> b" by auto
  with sup_trans1 have "y \<in> insert x X \<Longrightarrow> y \<le> sup b x" for y by (elim insertE,auto)
  then show ?case by auto
qed

end

class quasi_semilattice_sup = quasi_order_sup +
  assumes sup_least: "y \<le> x \<Longrightarrow> z \<le> x \<Longrightarrow> sup y z \<le> x"
begin

lemma finite_set_has_LUB: "finite X \<Longrightarrow> X \<noteq> {} \<Longrightarrow> LUB X \<noteq> {}"
proof(induct X rule:finite_induct)
  case empty then show ?case by auto
next
  case IH: (insert x X)
  show ?case
  proof(cases "X = {}")
    case True
    with LUB_singleton have "LUB (insert x X) = [x]\<^sub>\<sim>" by auto
    then show ?thesis by auto
  next
    case False
    with IH have "LUB X \<noteq> {}" by auto
    then obtain b where bLUB: "b \<in> LUB X" by auto
    have "sup x b \<in> LUB (insert x X)"
    proof
      from bLUB show "sup x b \<in> UB (insert x X)" by(auto intro: sup_trans2)
    next
      fix c assume cUB: "c \<in> UB (insert x X)"
      with bLUB show "sup x b \<le> c" by(auto intro: sup_least)
    qed
    then show ?thesis by auto
  qed
qed

end

subsection {* Products *}

lemma prod_less_cases[consumes 1]:
  assumes "x < y" and fst: "fst x < fst y \<Longrightarrow> P" and snd: "fst x \<le> fst y \<Longrightarrow> snd x < snd y \<Longrightarrow> P"
  shows P
  using assms[unfolded less_prod_def] by auto

lemma prod_less_eq_cases[consumes 1]:
  assumes "x \<le> y" and fst: "fst x < fst y \<Longrightarrow> P" and snd: "fst x \<le> fst y \<Longrightarrow> snd x \<le> snd y \<Longrightarrow> P"
  shows P
  using assms[unfolded less_eq_prod_def] by auto

lemma less_prodI1: "x < y \<Longrightarrow> (x,x') < (y,y')" by auto
lemma less_prodI2: "x \<le> y \<Longrightarrow> x' < y' \<Longrightarrow> (x,x') < (y,y')" by auto

lemma less_eq_prodI: "(fst x < fst y \<or> (fst x \<le> fst y \<and> snd x \<le> snd y)) \<Longrightarrow> x \<le> y"
  unfolding less_eq_prod_def by auto

instance prod :: (quasi_order,quasi_order) quasi_order
proof
  fix x y z :: "'a \<times> 'b"
  { assume "fst x < fst y" also assume "fst y < fst z" finally have "fst x < fst z". } note 1 = this
  { assume "fst x < fst y" also assume "fst y \<le> fst z" finally have "fst x < fst z". } note 2 = this
  { assume "fst x \<le> fst y" also assume "fst y < fst z" finally have "fst x < fst z". } note 3 = this
  { assume "fst x \<le> fst y" also assume "fst y \<le> fst z" finally have "fst x \<le> fst z". } note 4 = this
  { assume "snd x \<le> snd y" also assume "snd y \<le> snd z" finally have "snd x \<le> snd z". } note 5 = this
  { assume "snd x < snd y" also assume "snd y \<le> snd z" finally have "snd x < snd z". } note 6 = this
  { assume "snd x \<le> snd y" also assume "snd y < snd z" finally have "snd x < snd z". } note 7 = this
  { assume "snd x < snd y" also assume "snd y < snd z" finally have "snd x < snd z". } note 8 = this
  show "x \<ge> x" by (cases x,auto)
  from 1 2 3 4 5 show "x \<le> y \<Longrightarrow> y \<le> z \<Longrightarrow> x \<le> z" by (cases x, cases y, cases z, auto)
  from 1 2 3 4 6 show "x < y \<Longrightarrow> y \<le> z \<Longrightarrow> x < z" by (cases x, cases y, cases z, auto)
  from 1 2 3 4 7 show "x \<le> y \<Longrightarrow> y < z \<Longrightarrow> x < z" by (cases x, cases y, cases z, auto)
  from 1 2 3 4 8 show "x < y \<Longrightarrow> y < z \<Longrightarrow> x < z" by (cases x, cases y, cases z, auto)
qed

instance prod :: (wf_order, wf_order) wf_order
proof
  fix P :: "'a \<times> 'b \<Rightarrow> bool" and z :: "'a \<times> 'b"
  assume P: "\<And>x. (\<And>y. y < x \<Longrightarrow> P y) \<Longrightarrow> P x"
  show "P z"
  proof (induct z)
    case (Pair a b)
    show "P (a, b)"
    proof (induct a arbitrary: b rule: less_induct)
      case (less a\<^sub>1) note a\<^sub>1 = this
      have "a' \<le> a\<^sub>1 \<Longrightarrow> P (a', b)" for a'
      proof (induct b arbitrary: a' rule: less_induct)
        case (less b\<^sub>1) note b\<^sub>1 = this
        show "P (a', b\<^sub>1)"
        proof (rule P)
          fix p assume p: "p < (a', b\<^sub>1)"
          show "P p"
          proof (cases "fst p < a'")
            case True
            with b\<^sub>1(2) less_le_trans have "fst p < a\<^sub>1" by auto
            from a\<^sub>1[OF this] have "P (fst p, snd p)".
            then show ?thesis by simp
          next
            case False
            with p have 1: "fst p \<le> a'" and 2: "snd p < b\<^sub>1" by (simp_all add: less_prod_def)
            from 1 and b\<^sub>1(2) order_trans have "fst p \<le> a\<^sub>1" by auto
            from b\<^sub>1(1)[OF 2 this] have "P (fst p, snd p)".
            with 1 show ?thesis by (simp add: less_prod_def)
          qed
        qed
      qed
      then show "P (a\<^sub>1,b)" by auto
    qed
  qed
qed

section {* F-Algebras *}

locale algebra = fixes I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

fun interpret_term :: "('f,'v) term \<Rightarrow> ('v \<Rightarrow> 'a) \<Rightarrow> 'a" ("\<lbrakk>_\<rbrakk>(_)" [0,80]60)
  where "\<lbrakk>Var x\<rbrakk>\<alpha> = \<alpha> x"
      | "\<lbrakk>Fun f ss\<rbrakk>\<alpha> = I f (map (\<lambda>s. \<lbrakk>s\<rbrakk>\<alpha>) ss)"

definition subst_assign where "subst_assign \<sigma> \<alpha> x \<equiv> \<lbrakk>\<sigma> x\<rbrakk>\<alpha>"

lemma subst_assign: "\<lbrakk>s\<rbrakk>(subst_assign \<sigma> \<alpha>) = \<lbrakk>s\<cdot>\<sigma>\<rbrakk>\<alpha>"
proof(induct s)
  case (Var x) show ?case unfolding subst_assign_def by auto
next
  case (Fun f ss)
  { fix x note impI[of "x \<in> set ss", OF Fun[of x], simplified] }
  note * = map_ext[OF this]
  show ?case by (auto simp: comp_def *)
qed

end

locale term_algebra = algebra "Fun :: 'f \<Rightarrow> ('f,'v) term list \<Rightarrow> ('f,'v) term"
begin

lemma interpret_is_subst[simp]: "interpret_term s \<sigma> = s\<cdot>\<sigma>" by (induct s, unfold algebra.interpret_term.simps, auto)

lemma subst_assign_is_comp[simp]: "subst_assign \<sigma> \<tau> = \<sigma> \<circ>\<^sub>s \<tau>" by (auto simp: algebra.subst_assign_def subst_compose)

end

subsection {* Quasi-ordered algebra *}

text
{* A quasi-ordered algebra is a F-algebra whose domain is quasi-ordered.
   We cannot use @{class quasi_order} as a class since later we want to use
   arbitrary order on terms which cannot be globally fixed. Local instantiation
   would alleviate the problem.
 *}

locale quasi_ordered_algebra = quasi_order less_eq less + algebra I
  for less_eq :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

notation less_eq (infix "\<sqsubseteq>" 50)
notation less (infix "\<sqsubset>" 50)
abbreviation(input) greater_eq (infix "\<sqsupseteq>" 50) where "a \<sqsupseteq> b \<equiv> b \<sqsubseteq> a"
abbreviation(input) greater (infix "\<sqsupset>" 50) where "a \<sqsupset> b \<equiv> b \<sqsubset> a"

no_notation equiv (infix "\<sim>" 50)
definition equiv (infix "\<sim>" 50) where [iff]: "x \<sim> y \<equiv> x \<sqsubseteq> y \<and> y \<sqsubseteq> x"

no_notation equiv_class ("[_]\<^sub>\<sim>")
definition equiv_class ("[_]\<^sub>\<sim>") where "[x]\<^sub>\<sim> \<equiv> { y. x \<sim> y }"

text{* Base orderings are extended over terms in the standard way. *}

definition less_term (infix "\<sqsubset>\<^sub>t" 50) where "s \<sqsubset>\<^sub>t t \<equiv> \<forall>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>"
definition lesseq_term (infix "\<sqsubseteq>\<^sub>t" 50) where "s \<sqsubseteq>\<^sub>t t \<equiv> \<forall>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>"

abbreviation(input) greater_term (infix "\<sqsupset>\<^sub>t" 50) where "s \<sqsupset>\<^sub>t t \<equiv> t \<sqsubset>\<^sub>t s"
abbreviation(input) greatereq_term (infix "\<sqsupseteq>\<^sub>t" 50) where "s \<sqsupseteq>\<^sub>t t \<equiv> t \<sqsubseteq>\<^sub>t s"

definition equiv_term (infix "\<sim>\<^sub>t" 50) where [iff]: "s \<sim>\<^sub>t t \<equiv> \<forall>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sim> \<lbrakk>t\<rbrakk>\<alpha>"
definition equiv_class_term ("[_]\<^sub>t") where "[s]\<^sub>t \<equiv> { t. s \<sim>\<^sub>t t }"

lemma less_termI[intro]: "(\<And>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<sqsubset>\<^sub>t t" by (auto simp: less_term_def)
lemma less_termD[dest]: "s \<sqsubset>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>" by (auto simp: less_term_def)

lemma lesseq_termI[intro]: "(\<And>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<sqsubseteq>\<^sub>t t" by (auto simp: lesseq_term_def)
lemma lesseq_termD[dest]: "s \<sqsubseteq>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>" by (auto simp: lesseq_term_def)

lemma equiv_termI[intro]: "(\<And>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sim> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<sim>\<^sub>t t" by (auto simp: equiv_term_def)
lemma equiv_termD[dest]: "s \<sim>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sim> \<lbrakk>t\<rbrakk>\<alpha>" by (auto simp: equiv_term_def)

lemma equiv_term: "s \<sim>\<^sub>t t \<longleftrightarrow> s \<sqsubseteq>\<^sub>t t \<and> t \<sqsubseteq>\<^sub>t s" by fastforce

sublocale term_order: quasi_order "op \<sqsubseteq>\<^sub>t" "op \<sqsubset>\<^sub>t"
  apply (unfold_locales, unfold less_term_def lesseq_term_def)
  apply (auto dest: order_trans less_trans le_less_trans less_le_trans)
  done

lemma less_stable[intro]: "s \<sqsubset>\<^sub>t t \<Longrightarrow> s\<cdot>\<sigma> \<sqsubset>\<^sub>t t\<cdot>\<sigma>" by (intro less_termI, fold subst_assign, auto)
lemma lesseq_stable[intro]: "s \<sqsubseteq>\<^sub>t t \<Longrightarrow> s\<cdot>\<sigma> \<sqsubseteq>\<^sub>t t\<cdot>\<sigma>" by (intro lesseq_termI, fold subst_assign, auto)

end


locale wf_algebra = quasi_ordered_algebra less_eq less I + wf_order less_eq less
  for less_eq and less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

sublocale term_order: wf_order "op \<sqsubseteq>\<^sub>t :: ('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" "op \<sqsubset>\<^sub>t"
proof
  obtain v where "(v :: 'a) = v" by auto
  def \<alpha> \<equiv> "\<lambda>x :: 'v. v"
  fix P s
  assume *: "\<And>s :: ('f,'v) term. (\<And>t. t \<sqsubset>\<^sub>t s \<Longrightarrow> P t) \<Longrightarrow> P s"
  then show "P s"
  proof(induct "\<lbrakk>s\<rbrakk>\<alpha>" arbitrary: s rule: less_induct)
    case less show ?case
    proof (rule less(2))
      fix t assume ts: "t \<sqsubset>\<^sub>t s"
      then have "\<lbrakk>t\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>s\<rbrakk>\<alpha>" by auto
      from less(1)[OF this *] show "P t" by auto
    qed
  qed
qed

end

lemma(in quasi_order) weak_mono_iff_all:
  "(\<forall>a b ls rs. a \<le> b \<longrightarrow> f (ls @ a # rs) \<le> f (ls @ b # rs)) \<longleftrightarrow>
   (\<forall>as bs. length as = length bs \<longrightarrow> (\<forall>i < length as. as ! i \<le> bs ! i) \<longrightarrow> f as \<le> f bs)"
  (is "?l \<longleftrightarrow> ?r")
proof(intro iffI impI allI)
  fix as bs
  assume mono: ?l and all: "\<forall>i < length as. as ! i \<le> bs ! i" and len: "length as = length bs"
  { fix cs
    from all have "\<And>i. i < length as \<Longrightarrow> as ! i \<le> bs ! i" by auto
    with len have "f (cs @ as) \<le> f (cs @ bs)"
    proof(induct as arbitrary: bs cs)
      case Nil then have "bs = []" by auto
      with order_refl show ?case by auto
    next
      case IH: (Cons a as)
      then have "Suc (length as) = length bs" by auto
      then obtain d ds where bs: "bs = d # ds" and len: "length as = length ds" by (induct bs, auto)
      have "f ((cs @ [a]) @ as) \<le> f ((cs @ [a]) @ ds)"
        apply(rule IH(1)[OF len]) using IH(3) bs by fastforce
      also have "... \<le> f ((cs @ [d]) @ ds)"
        proof-
          from IH(3)[unfolded bs, of 0] have "a \<le> d" by auto
          with mono show ?thesis by auto
        qed
      finally show ?case by (unfold bs, auto)
    qed
  }
  from this[of "[]"] show "f as \<le> f bs" by auto
next
  assume ?r
  then have *: "\<And>as bs. length as = length bs \<Longrightarrow> (\<forall>i<length as. as ! i \<le> bs ! i) \<Longrightarrow> f as \<le> f bs"
    by auto
  fix a b ls rs
  assume ab: "a \<le> b"
  show "f (ls @ a # rs) \<le> f (ls @ b # rs)"
  proof (rule *,simp, intro allI impI)
    fix i assume i: "i < length (ls @ a # rs)"
    show " (ls @ a # rs) ! i \<le> (ls @ b # rs) ! i"
    by(cases i "length ls" rule: linorder_cases, auto simp: ab nth_append)
  qed
qed

locale weak_mono_fun = quasi_order less_eq less
  for less_eq less +
  fixes f
  assumes mono: "\<And>ls rs a b. less_eq a b \<Longrightarrow> less_eq (f (ls @ a # rs)) (f (ls @ b # rs))"
begin

lemma mono_all:
  assumes "length as = length bs"
    and "\<And>i. i < length as \<Longrightarrow> less_eq (as ! i) (bs ! i)"
  shows "less_eq (f as) (f bs)"
proof-
  from mono
  have "\<forall>as bs. length as = length bs \<longrightarrow> (\<forall>i<length as. less_eq (as ! i) (bs ! i)) \<longrightarrow> less_eq (f as) (f bs)"
    by(subst weak_mono_iff_all[symmetric], auto)
  with assms show ?thesis by auto
qed

end

subsection {* Weakly monotone algebra *}

locale weak_mono_algebra = quasi_ordered_algebra less_eq less I
  for less_eq and less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  assumes weak_mono: "\<And>f ls rs a b. less_eq a b \<Longrightarrow> less_eq (I f (ls @ a # rs)) (I f (ls @ b # rs))"
begin

sublocale weak_mono_fun less_eq less "I f" by(unfold_locales, fact weak_mono)

interpretation weak_mono_term: weak_mono_fun lesseq_term less_term "Fun f"
  by (unfold_locales, intro lesseq_termI, auto intro!: weak_mono)

lemmas weak_mono_term = weak_mono_term.mono
lemmas weak_mono_term_all = weak_mono_term.mono_all

lemma mono_range[intro]: "(\<And>v. v \<in> vars_term s \<Longrightarrow> \<alpha> v \<sqsubseteq> \<beta> v) \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>s\<rbrakk>\<beta>"
proof (induct s)
  case (Var v) then show ?case by auto
next
  case IH: (Fun f ss)
  { fix i assume "i < length ss"
    moreover
      then have "vars_term (ss!i) \<subseteq> vars_term (Fun f ss)" by fastforce
      with IH(2) have *: "v \<in> vars_term (ss!i) \<Longrightarrow> \<alpha> v \<sqsubseteq> \<beta> v" for v by blast
    ultimately have "\<lbrakk>ss ! i\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>ss ! i\<rbrakk>\<beta>" by(intro IH(1), auto)
  }
  then show ?case by(auto intro: mono_all)
qed

lemma ctxt_closed_NS: "ctxt.closed term_order.NS"
 by (rule one_imp_ctxt_closed, auto intro: weak_mono_term.mono simp: term_order.NS_def)
lemma subst_closed_S: "subst.closed term_order.S" by (unfold term_order.S_def, auto)
lemma subst_closed_NS: "subst.closed term_order.NS" by (unfold term_order.NS_def, auto)

lemma lesseq_ctxt_closed[intro]: "s \<sqsupseteq>\<^sub>t t \<Longrightarrow> C\<langle>s\<rangle> \<sqsupseteq>\<^sub>t C\<langle>t\<rangle>"
  using ctxt.closedD[OF ctxt_closed_NS] by (auto simp: term_order.NS_def)

end

locale weak_mono_wf_algebra = weak_mono_algebra less_eq less I + wf_order less_eq less
  for less_eq less I
begin

sublocale wf_algebra..

sublocale redpair: redpair "term_order.S" "term_order.NS"
  using SN ctxt_closed_NS subst_closed_S subst_closed_NS by unfold_locales

end

locale mono_algebra = weak_mono_algebra less_eq less I for less_eq less I +
  assumes mono: "\<And>f ls rs a b. less a b \<Longrightarrow> less (I f (ls @ a # rs)) (I f (ls @ b # rs))"
begin

lemma mono_term: "s \<sqsubset>\<^sub>t t \<Longrightarrow> Fun f (ls @ s # rs) \<sqsubset>\<^sub>t Fun f (ls @ t # rs)"
  by(intro less_termI, auto intro: mono)

lemma ctxt_closed_S: "ctxt.closed term_order.S"
  by (rule one_imp_ctxt_closed, auto intro: mono_term simp: term_order.S_def)

lemma less_ctxt_closed[intro]: "s \<sqsupset>\<^sub>t t \<Longrightarrow> C\<langle>s\<rangle> \<sqsupset>\<^sub>t C\<langle>t\<rangle>"
  using ctxt.closedD[OF ctxt_closed_S] by (auto simp: term_order.S_def)

end


section {* Models *}

locale quasi_model = weak_mono_algebra less_eq less I
  for less_eq less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  fixes R :: "('f,'v) trs"
  assumes compat: "(l,r) \<in> R \<Longrightarrow> l \<sqsupseteq>\<^sub>t r"
begin

lemma rstep_imp_lesseq: "(s,t) \<in> (rstep R)\<^sup>* \<Longrightarrow> s \<sqsupseteq>\<^sub>t t"
proof(induct rule: rtrancl_induct)
  case base show ?case by auto
next
  case (step t u)
  then obtain C \<sigma> l r where t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>" and u: "u = C\<langle>r\<cdot>\<sigma>\<rangle>" and lr: "(l,r) \<in> R" by auto
  have "t \<sqsupseteq>\<^sub>t u" by (unfold t u, auto intro: compat lr)
  also note step(3)
  finally show ?case.
qed

end

locale equimodel = weak_mono_algebra less_eq less I
  for less_eq less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  fixes E :: "('f,'v) trs"
  assumes compat: "(l,r) \<in> E \<Longrightarrow> l \<sim>\<^sub>t r"
begin

sublocale quasi_model less_eq less I "E\<^sup>\<leftrightarrow>"
  by (unfold_locales, insert compat[unfolded equiv_term_def], auto)

lemma conversion_imp_equiv: assumes "(s,t) \<in> (rstep E)\<^sup>\<leftrightarrow>\<^sup>*" shows "s \<sim>\<^sub>t t"
proof-
  note * = rstep_imp_lesseq[unfolded rstep_simps(5), folded conversion_def]
  from *[OF assms[unfolded conversion_inv[of s t]]] *[OF assms]
  show ?thesis by (auto simp: equiv_term)
qed

end

locale model = equimodel less_eq less I E + weak_order less_eq less for less_eq less I E
begin

lemma conversion_imp_eq: assumes "(s,t) \<in> (rstep E)\<^sup>\<leftrightarrow>\<^sup>*" shows "\<lbrakk>s\<rbrakk>\<alpha> = \<lbrakk>t\<rbrakk>\<alpha>"
  using conversion_imp_equiv[OF assms] by auto

end

subsection {* Encoding algebras *}

text {* Encoding $F$-algebra into $G$-algebra.
 Each function symbol $f \in F$ of arity $n$ is encoded as a term in $T(G,{0..<n})$. *}

locale encoding =
  fixes E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
  assumes var_domain: "i \<in> vars_term (E (f,n)) \<Longrightarrow> i < n"
begin

end


locale encoded_algebra = base: algebra I + encoding E
  for I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

definition IE :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
  where "IE f vs = base.interpret_term (E (f, length vs)) (nth vs)"

sublocale algebra IE.

end

locale encoded_quasi_ordered_algebra =
  quasi_order less_eq less + encoding E
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

sublocale base: quasi_ordered_algebra less_eq less I..

sublocale encoded_algebra I E..

sublocale quasi_ordered_algebra less_eq less IE..

end

locale encoded_wf_algebra =
  wf_order less_eq less + encoding E
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

sublocale base: wf_algebra less_eq less I..

sublocale encoded_quasi_ordered_algebra..

sublocale wf_algebra less_eq less IE..

end


locale encoded_weak_mono_algebra =
  quasi_order less_eq less + encoding E
  for less_eq less
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term" +
  assumes mono: "\<And>f ls rs a b. less_eq a b \<Longrightarrow> less_eq (I f (ls @ a # rs)) (I f (ls @ b # rs))"

begin

sublocale base: weak_mono_algebra less_eq less I by (unfold_locales, fact mono)

sublocale encoded_quasi_ordered_algebra less_eq less I E..

sublocale weak_mono_algebra less_eq less IE
proof (unfold_locales)
  fix f ls rs a b
  { fix as bs :: "'a list"
    assume *: "length as = length bs"
    with var_domain[of _ f]
    have "(\<forall>i < length as. as ! i \<sqsubseteq> bs ! i) \<Longrightarrow> IE f as \<sqsubseteq> IE f bs" by (auto simp: IE_def)
  }
  note * = this
  assume ab: "a \<sqsubseteq> b"
  show "IE f (ls @ a # rs) \<sqsubseteq> IE f (ls @ b # rs)"
  proof(intro * allI impI, simp)
    fix i assume "i < length (ls @ a # rs)"
    from ab show "(ls @ a # rs) ! i \<sqsubseteq> (ls @ b # rs) ! i"
    by(cases i "length ls" rule: linorder_cases, auto simp: nth_append)
  qed
qed

end

locale encoded_weak_mono_wf_algebra =
  wf_order less_eq less + encoded_weak_mono_algebra less_eq less I E for less_eq less I E
begin

sublocale encoded_wf_algebra less_eq less I E..

sublocale weak_mono_wf_algebra less_eq less IE..

end

section {* Max-polynomial Interpretations as an Instance *}

text {* As the base algebra, we introduce an $G$-algebra where $G = \{+,\times,\max\}$. *}

locale maxpoly begin

datatype sig = Fconst nat | Fsum | Fprod | Fmax

fun I where
  "I (Fconst n) = (\<lambda>x. n)"
| "I Fsum = listsum"
| "I Fprod = listprod"
| "I Fmax = Max \<circ> set"

sublocale weak_mono_wf_algebra less_eq less "I :: _ \<Rightarrow> nat list \<Rightarrow> nat"
proof
  fix f ls rs and a b :: nat
  assume ba: "b \<le> a"
  then show "I f (ls @ b # rs) \<le> I f (ls @ a # rs)"
  proof(cases f)
    case Fmax
    note ba
    also have "a \<le> Max (insert a (set ls \<union> set rs))" by auto
    finally show ?thesis by (auto simp: Fmax)
  qed auto
qed

end

text {* Now we encode arbitrary $F$-terms into the $G$-terms. *}

locale maxpoly_interpretation = encoding E
  for E :: "'f \<times> nat \<Rightarrow> (maxpoly.sig, nat) term"
begin

sublocale maxpoly: maxpoly.

sublocale encoded_weak_mono_wf_algebra "less_eq::nat\<Rightarrow>nat\<Rightarrow>bool" less maxpoly.I E
  by (unfold_locales, fact "maxpoly.mono")

text {* Now we get a reduction pair for free! *}

sublocale redpair: redpair "term_order.S" "term_order.NS"..

end


















end